# 著作者：叶凯文
# Author: YE KAIWEN
# 版权归著作者叶凯文所有
# Copyright (c) 2022 YE KAIWEN. All Rights Reserved.

import pandas as pd

class NormalTrade:
    def __init__(self,buyfee,sellfee,stopprofit=None,stoploss=None):
        self.buyfee=buyfee
        self.sellfee=sellfee
        self.extralogics=[]
        if stopprofit:
            extralogic=stopprofit_logic(stopprofit=stopprofit)
            self.extralogics.append(extralogic)
        if stoploss:
            extralogic=stoploss_logic(stoploss=stoploss)
            self.extralogics.append(extralogic)

    def statistics(self,data):
        if (set(['orderbuy','ordersell','refer','high','low','openlogic','holdlogic'])<=set(data.columns))==False:
            print('缺少必要数据列',set(['orderbuy','ordersell','refer','high','low','openlogic','holdlogic']).difference(set(data.columns)))
            return None
        status='close'
        trades=[]
        for inx,row in data.iterrows():
            if status=='close':
                if (row['openlogic']&row['holdlogic']):
                    status='buying'
                    orderday=0
            elif status=='buying':
                orderday+=1
                if (row['orderbuy']>=row['low']):
                    status='open'
                    orderinfo={'buyinx':inx,'buyday':orderday,'buybid':min(row['orderbuy'],row['high'])}
                if status=='open':
                    revbase=row['refer']
                    revenue=(revbase/orderinfo['buybid'])*(1-self.sellfee)/(1+self.buyfee)-1
                    hold = row['holdlogic']
                    for logic in self.extralogics:
                        hold = hold & logic.calc(revenue)
                    if hold==False:
                        status = 'selling'
                        orderday = 0
                else:
                    if (row['openlogic']&row['holdlogic'])==False:
                        status='close'
            elif status=='open':
                revbase = row['refer']
                revenue = (revbase / orderinfo['buybid']) * (1 - self.sellfee)/(1 + self.buyfee) - 1
                hold = row['holdlogic']
                for logic in self.extralogics:
                    hold = hold & logic.calc(revenue)
                if hold == False:
                    status = 'selling'
                    orderday = 0
            elif status=='selling':
                orderday+=1
                if row['ordersell']<=row['high']:
                    orderinfo['sellinx']=inx
                    orderinfo['sellday']=orderday
                    orderinfo['sellbid']=max(row['ordersell'],row['low'])
                    orderinfo['revenue']=(orderinfo['sellbid']/orderinfo['buybid'])*(1-self.sellfee)/(1+self.buyfee)-1
                    orderinfo['cost']=(1-1/(1+self.buyfee))+(orderinfo['sellbid']/orderinfo['buybid'])*self.sellfee/(1+self.buyfee)
                    highpiece=data['high'][orderinfo['buyinx']:orderinfo['sellinx']]
                    lowpiece = data['low'][orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=data['ordersell'][orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=(highpiece>=sppiece)*sppiece+(highpiece<sppiece)*highpiece
                    sppiece=(lowpiece<=sppiece)*sppiece+(lowpiece>sppiece)*lowpiece
                    spmax=sppiece.max()
                    spmin=sppiece.min()
                    orderinfo['revmax']=(spmax/orderinfo['buybid'])*(1-self.sellfee)/(1+self.buyfee)-1
                    orderinfo['revmin']=(spmin/orderinfo['buybid'])*(1-self.sellfee)/(1+self.buyfee)-1
                    trades.append(orderinfo)
                    status='close'
                if status=='close':
                    if (row['openlogic'] & row['holdlogic']):
                        status = 'buying'
                        orderday = 0
                else:
                    revbase = row['refer']
                    revenue = (revbase / orderinfo['buybid']) * (1 - self.sellfee)/(1+self.buyfee) - 1
                    hold = row['holdlogic']
                    for logic in self.extralogics:
                        hold = hold & logic.calc(revenue)
                    if hold:
                        status = 'open'
            else:
                raise Exception("Illegal status!")
        if status!='close':
            orderinfo['status']=status
            trades.append(orderinfo)
        return trades

    def statistics_with_callauction(self,data):
        if (set(['conbuy','consell','callbuy','callsell','refer','open','high','low','openlogic','holdlogic'])<=set(data.columns))==False:
            print('缺少必要数据列',set(['conbuy','consell','callbuy','callsell','refer','open','high','low','openlogic','holdlogic']).difference(set(data.columns)))
            return None
        status='close'
        trades=[]
        for inx,row in data.iterrows():
            if status=='close':
                if (row['openlogic']&row['holdlogic']):
                    status='buying'
                    orderday=0
            elif status=='buying':
                orderday+=1
                if (row['callbuy']>=row['open']):
                    status='open'
                    orderinfo={'buyinx':inx,'buyday':orderday,'buybid':row['open']}
                elif (row['conbuy']>=row['low']):
                    status='open'
                    orderinfo={'buyinx':inx,'buyday':orderday,'buybid':min(row['conbuy'],row['high'])}
                if status=='open':
                    revbase=row['refer']
                    revenue=(revbase/orderinfo['buybid'])*(1-self.sellfee)/(1+self.buyfee)-1
                    hold = row['holdlogic']
                    for logic in self.extralogics:
                        hold = hold & logic.calc(revenue)
                    if hold==False:
                        status = 'selling'
                        orderday = 0
                else:
                    if (row['openlogic']&row['holdlogic'])==False:
                        status='close'
            elif status=='open':
                revbase = row['refer']
                revenue = (revbase / orderinfo['buybid']) * (1 - self.sellfee)/(1+self.buyfee) - 1
                hold = row['holdlogic']
                for logic in self.extralogics:
                    hold = hold & logic.calc(revenue)
                if hold == False:
                    status = 'selling'
                    orderday = 0
            elif status=='selling':
                orderday+=1
                if row['callsell']<=row['open']:
                    orderinfo['sellinx']=inx
                    orderinfo['sellday']=orderday
                    orderinfo['sellbid']=row['open']
                    orderinfo['revenue']=(orderinfo['sellbid']/orderinfo['buybid'])*(1-self.sellfee)/(1+self.buyfee)-1
                    orderinfo['cost']=(1-1/(1+self.buyfee))+(orderinfo['sellbid']/orderinfo['buybid'])*self.sellfee/(1+self.buyfee)
                    highpiece=data['high'][orderinfo['buyinx']:orderinfo['sellinx']]
                    lowpiece = data['low'][orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=data['refer'].shift()[orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=(highpiece>=sppiece)*sppiece+(highpiece<sppiece)*highpiece
                    sppiece=(lowpiece<=sppiece)*sppiece+(lowpiece>sppiece)*lowpiece
                    spmax=sppiece.max()
                    spmin=sppiece.min()
                    orderinfo['revmax']=(spmax/orderinfo['buybid'])*(1-self.sellfee)/(1+self.buyfee)-1
                    orderinfo['revmin']=(spmin/orderinfo['buybid'])*(1-self.sellfee)/(1+self.buyfee)-1
                    trades.append(orderinfo)
                    status='close'
                elif row['consell']<=row['high']:
                    orderinfo['sellinx']=inx
                    orderinfo['sellday']=orderday
                    orderinfo['sellbid']=max(row['consell'],row['low'])
                    orderinfo['revenue']=(orderinfo['sellbid']/orderinfo['buybid'])*(1-self.sellfee)/(1+self.buyfee)-1
                    orderinfo['cost']=(1-1/(1+self.buyfee))+(orderinfo['sellbid']/orderinfo['buybid'])*self.sellfee/(1+self.buyfee)
                    highpiece=data['high'][orderinfo['buyinx']:orderinfo['sellinx']]
                    lowpiece = data['low'][orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=data['refer'].shift()[orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=(highpiece>=sppiece)*sppiece+(highpiece<sppiece)*highpiece
                    sppiece=(lowpiece<=sppiece)*sppiece+(lowpiece>sppiece)*lowpiece
                    spmax=sppiece.max()
                    spmin=sppiece.min()
                    orderinfo['revmax']=(spmax/orderinfo['buybid'])*(1-self.sellfee)/(1+self.buyfee)-1
                    orderinfo['revmin']=(spmin/orderinfo['buybid'])*(1-self.sellfee)/(1+self.buyfee)-1
                    trades.append(orderinfo)
                    status='close'
                if status=='close':
                    if (row['openlogic'] & row['holdlogic']):
                        status = 'buying'
                        orderday = 0
                else:
                    revbase = row['refer']
                    revenue =  (revbase / orderinfo['buybid']) * (1 - self.sellfee)/(1+self.buyfee) - 1
                    hold = row['holdlogic']
                    for logic in self.extralogics:
                        hold = hold & logic.calc(revenue)
                    if hold:
                        status = 'open'
            else:
                raise Exception("Illegal status!")
        if status!='close':
            orderinfo['status']=status
            trades.append(orderinfo)
        return trades

class MarginTrade:
    def __init__(self,buyfee,sellfee,interest,margin,guarantee,stopprofit=None,stoploss=None):
        self.buyfee=buyfee
        self.sellfee=sellfee
        self.interest=interest/360
        self.margin=margin
        self.guarantee=guarantee
        self.extralogics = []
        if stopprofit:
            extralogic=stopprofit_logic(stopprofit=stopprofit)
            self.extralogics.append(extralogic)
        if stoploss:
            extralogic=stoploss_logic(stoploss=stoploss)
            self.extralogics.append(extralogic)

    def statistics(self,data):
        if (set(['orderbuy','ordersell','closeout','refer','high','low','openlogic','holdlogic'])<=set(data.columns))==False:
            print('缺少必要数据列',set(['orderbuy','ordersell','closeout','refer','high','low','openlogic','holdlogic']).difference(set(data.columns)))
            return None
        status='close'
        trades=[]
        for inx,row in data.iterrows():
            if status=='close':
                if (row['openlogic']&row['holdlogic']):
                    status='buying'
                    orderday=0
            elif status=='buying':
                orderday+=1
                if (row['orderbuy']>=row['low']):
                    status='open'
                    holdday=0
                    orderinfo={'buyinx':inx,'buyday':orderday,'buybid':min(row['orderbuy'],row['high'])}
                if status=='open':
                    revbase=row['refer']
                    revenue=((revbase/orderinfo['buybid'])*(1-self.sellfee)-(1+self.interest*holdday))/(self.margin+self.buyfee)
                    guarantee=(revbase/orderinfo['buybid'])*(1-self.sellfee)-self.interest*holdday+self.margin
                    if guarantee<self.guarantee:
                        status='mco'
                        orderday=0
                    else:
                        hold = row['holdlogic']
                        for logic in self.extralogics:
                            hold = hold & logic.calc(revenue)
                        if hold==False:
                            status = 'selling'
                            orderday = 0
                else:
                    if (row['openlogic']&row['holdlogic'])==False:
                        status='close'
            elif status=='open':
                holdday+=1
                revbase = row['refer']
                revenue = ((revbase / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * holdday)) / (self.margin + self.buyfee)
                guarantee = (revbase / orderinfo['buybid']) * (1 - self.sellfee) - self.interest * holdday + self.margin
                if guarantee < self.guarantee:
                    status = 'mco'
                    orderday = 0
                else:
                    hold = row['holdlogic']
                    for logic in self.extralogics:
                        hold = hold & logic.calc(revenue)
                    if hold == False:
                        status = 'selling'
                        orderday = 0
            elif status=='mco':
                holdday += 1
                orderday +=1
                if row['closeout']<=row['high']:
                    orderinfo['sellinx']=inx
                    orderinfo['sellday']=orderday
                    orderinfo['sellbid']=max(row['closeout'],row['low'])
                    orderinfo['revenue']=((orderinfo['sellbid'] / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * holdday)) / (self.margin + self.buyfee)
                    orderinfo['guarantee']=(orderinfo['sellbid'] / orderinfo['buybid']) * (1 - self.sellfee) - self.interest * holdday + self.margin
                    orderinfo['cost']=(self.buyfee+(orderinfo['sellbid'] / orderinfo['buybid'])*self.sellfee+self.interest * holdday)/(self.margin + self.buyfee)
                    orderinfo['closetype']='margin-closeout'
                    highpiece=data['high'][orderinfo['buyinx']:orderinfo['sellinx']]
                    lowpiece = data['low'][orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=data['ordersell'][orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=(highpiece>=sppiece)*sppiece+(highpiece<sppiece)*highpiece
                    sppiece=(lowpiece<=sppiece)*sppiece+(lowpiece>sppiece)*lowpiece
                    spdays=pd.Series(range(len(sppiece)),index=sppiece.index)
                    revenueseries=((sppiece / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * spdays)) / (self.margin + self.buyfee)
                    orderinfo['revmax']=revenueseries.max()
                    orderinfo['revmin']=revenueseries.min()
                    trades.append(orderinfo)
                    status='close'
            elif status=='selling':
                holdday += 1
                orderday+=1
                if row['ordersell']<=row['high']:
                    orderinfo['sellinx']=inx
                    orderinfo['sellday']=orderday
                    orderinfo['sellbid']=max(row['ordersell'],row['low'])
                    orderinfo['revenue']=((orderinfo['sellbid'] / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * holdday)) / (self.margin + self.buyfee)
                    orderinfo['guarantee']=(orderinfo['sellbid'] / orderinfo['buybid']) * (1 - self.sellfee) - self.interest * holdday + self.margin
                    orderinfo['cost']=(self.buyfee+(orderinfo['sellbid'] / orderinfo['buybid'])*self.sellfee+self.interest * holdday)/(self.margin + self.buyfee)
                    orderinfo['closetype']='sell-close'
                    highpiece=data['high'][orderinfo['buyinx']:orderinfo['sellinx']]
                    lowpiece = data['low'][orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=data['ordersell'][orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=(highpiece>=sppiece)*sppiece+(highpiece<sppiece)*highpiece
                    sppiece=(lowpiece<=sppiece)*sppiece+(lowpiece>sppiece)*lowpiece
                    spdays = pd.Series(range(len(sppiece)), index=sppiece.index)
                    revenueseries = ((sppiece / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * spdays)) / (self.margin + self.buyfee)
                    orderinfo['revmax'] = revenueseries.max()
                    orderinfo['revmin'] = revenueseries.min()
                    trades.append(orderinfo)
                    status='close'
                if status=='close':
                    if (row['openlogic'] & row['holdlogic']):
                        status = 'buying'
                        orderday = 0
                else:
                    revbase = row['refer']
                    revenue = ((revbase / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * holdday)) / (self.margin + self.buyfee)
                    guarantee = (revbase / orderinfo['buybid']) * (1 - self.sellfee) - self.interest * holdday + self.margin
                    if guarantee < self.guarantee:
                        status = 'mco'
                        orderday = 0
                    else:
                        hold = row['holdlogic']
                        for logic in self.extralogics:
                            hold = hold & logic.calc(revenue)
                        if hold:
                            status = 'open'
            else:
                raise Exception("Illegal status!")
        if status!='close':
            orderinfo['status']=status
            trades.append(orderinfo)
        return trades

    def statistics_with_callauction(self,data):
        if (set(['conbuy','consell','closeout','callbuy','callsell','refer','high','low','openlogic','holdlogic'])<=set(data.columns))==False:
            print('缺少必要数据列',set(['conbuy','consell','closeout','callbuy','callsell','refer','high','low','openlogic','holdlogic']).difference(set(data.columns)))
            return None
        status='close'
        trades=[]
        for inx,row in data.iterrows():
            if status=='close':
                if (row['openlogic']&row['holdlogic']):
                    status='buying'
                    orderday=0
            elif status=='buying':
                orderday+=1
                if (row['callbuy']>=row['open']):
                    status='open'
                    holdday=0
                    orderinfo={'buyinx':inx,'buyday':orderday,'buybid':row['open']}
                elif (row['conbuy']>=row['low']):
                    status='open'
                    holdday=0
                    orderinfo={'buyinx':inx,'buyday':orderday,'buybid':min(row['conbuy'],row['high'])}
                if status=='open':
                    revbase=row['refer']
                    revenue=((revbase/orderinfo['buybid'])*(1-self.sellfee)-(1+self.interest*holdday))/(self.margin+self.buyfee)
                    guarantee=(revbase/orderinfo['buybid'])*(1-self.sellfee)-self.interest*holdday+self.margin
                    if guarantee<self.guarantee:
                        status='mco'
                        orderday=0
                    else:
                        hold = row['holdlogic']
                        for logic in self.extralogics:
                            hold = hold & logic.calc(revenue)
                        if hold==False:
                            status = 'selling'
                            orderday = 0
                else:
                    if (row['openlogic']&row['holdlogic'])==False:
                        status='close'
            elif status=='open':
                holdday+=1
                revbase = row['refer']
                revenue = ((revbase / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * holdday)) / (self.margin + self.buyfee)
                guarantee = (revbase / orderinfo['buybid']) * (1 - self.sellfee) - self.interest * holdday + self.margin
                if guarantee < self.guarantee:
                    status = 'mco'
                    orderday = 0
                else:
                    hold = row['holdlogic']
                    for logic in self.extralogics:
                        hold = hold & logic.calc(revenue)
                    if hold == False:
                        status = 'selling'
                        orderday = 0
            elif status=='mco':
                holdday += 1
                orderday +=1
                if row['closeout']<=row['open']:
                    orderinfo['sellinx'] = inx
                    orderinfo['sellday'] = orderday
                    orderinfo['sellbid'] = row['open']
                    orderinfo['revenue'] = ((orderinfo['sellbid'] / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * holdday)) / (self.margin + self.buyfee)
                    orderinfo['guarantee'] = (orderinfo['sellbid'] / orderinfo['buybid']) * (1 - self.sellfee) - self.interest * holdday + self.margin
                    orderinfo['cost'] = (self.buyfee + (orderinfo['sellbid'] / orderinfo['buybid']) * self.sellfee + self.interest * holdday) / (self.margin + self.buyfee)
                    orderinfo['closetype'] = 'margin-closeout'
                    highpiece = data['high'][orderinfo['buyinx']:orderinfo['sellinx']]
                    lowpiece = data['low'][orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece = data['refer'].shift()[orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece = (highpiece >= sppiece) * sppiece + (highpiece < sppiece) * highpiece
                    sppiece = (lowpiece <= sppiece) * sppiece + (lowpiece > sppiece) * lowpiece
                    spdays = pd.Series(range(len(sppiece)), index=sppiece.index)
                    revenueseries = ((sppiece / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * spdays)) / (self.margin + self.buyfee)
                    orderinfo['revmax'] = revenueseries.max()
                    orderinfo['revmin'] = revenueseries.min()
                    trades.append(orderinfo)
                    status = 'close'
                elif row['closeout']<=row['high']:
                    orderinfo['sellinx']=inx
                    orderinfo['sellday']=orderday
                    orderinfo['sellbid']=max(row['closeout'],row['low'])
                    orderinfo['revenue']=((orderinfo['sellbid'] / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * holdday)) / (self.margin + self.buyfee)
                    orderinfo['guarantee']=(orderinfo['sellbid'] / orderinfo['buybid']) * (1 - self.sellfee) - self.interest * holdday + self.margin
                    orderinfo['cost']=(self.buyfee+(orderinfo['sellbid'] / orderinfo['buybid'])*self.sellfee+self.interest * holdday)/(self.margin + self.buyfee)
                    orderinfo['closetype']='margin-closeout'
                    highpiece=data['high'][orderinfo['buyinx']:orderinfo['sellinx']]
                    lowpiece = data['low'][orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=data['refer'].shift()[orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=(highpiece>=sppiece)*sppiece+(highpiece<sppiece)*highpiece
                    sppiece=(lowpiece<=sppiece)*sppiece+(lowpiece>sppiece)*lowpiece
                    spdays = pd.Series(range(len(sppiece)), index=sppiece.index)
                    revenueseries = ((sppiece / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * spdays)) / (self.margin + self.buyfee)
                    orderinfo['revmax'] = revenueseries.max()
                    orderinfo['revmin'] = revenueseries.min()
                    trades.append(orderinfo)
                    status='close'
            elif status=='selling':
                holdday += 1
                orderday+=1
                if row['callsell']<=row['open']:
                    orderinfo['sellinx']=inx
                    orderinfo['sellday']=orderday
                    orderinfo['sellbid']=row['open']
                    orderinfo['revenue']=((orderinfo['sellbid'] / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * holdday)) / (self.margin + self.buyfee)
                    orderinfo['guarantee']=(orderinfo['sellbid'] / orderinfo['buybid']) * (1 - self.sellfee) - self.interest * holdday + self.margin
                    orderinfo['cost']=(self.buyfee+(orderinfo['sellbid'] / orderinfo['buybid'])*self.sellfee+self.interest * holdday)/(self.margin + self.buyfee)
                    orderinfo['closetype']='sell-close'
                    highpiece=data['high'][orderinfo['buyinx']:orderinfo['sellinx']]
                    lowpiece = data['low'][orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=data['refer'].shift()[orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=(highpiece>=sppiece)*sppiece+(highpiece<sppiece)*highpiece
                    sppiece=(lowpiece<=sppiece)*sppiece+(lowpiece>sppiece)*lowpiece
                    spdays = pd.Series(range(len(sppiece)), index=sppiece.index)
                    revenueseries = ((sppiece / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * spdays)) / (self.margin + self.buyfee)
                    orderinfo['revmax'] = revenueseries.max()
                    orderinfo['revmin'] = revenueseries.min()
                    trades.append(orderinfo)
                    status='close'
                elif row['consell']<=row['high']:
                    orderinfo['sellinx']=inx
                    orderinfo['sellday']=orderday
                    orderinfo['sellbid']=max(row['consell'],row['low'])
                    orderinfo['revenue']=((orderinfo['sellbid'] / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * holdday)) / (self.margin + self.buyfee)
                    orderinfo['guarantee']=(orderinfo['sellbid'] / orderinfo['buybid']) * (1 - self.sellfee) - self.interest * holdday + self.margin
                    orderinfo['cost']=(self.buyfee+(orderinfo['sellbid'] / orderinfo['buybid'])*self.sellfee+self.interest * holdday)/(self.margin + self.buyfee)
                    orderinfo['closetype']='sell-close'
                    highpiece=data['high'][orderinfo['buyinx']:orderinfo['sellinx']]
                    lowpiece = data['low'][orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=data['refer'].shift()[orderinfo['buyinx']:orderinfo['sellinx']]
                    sppiece=(highpiece>=sppiece)*sppiece+(highpiece<sppiece)*highpiece
                    sppiece=(lowpiece<=sppiece)*sppiece+(lowpiece>sppiece)*lowpiece
                    spdays = pd.Series(range(len(sppiece)), index=sppiece.index)
                    revenueseries = ((sppiece / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * spdays)) / (self.margin + self.buyfee)
                    orderinfo['revmax'] = revenueseries.max()
                    orderinfo['revmin'] = revenueseries.min()
                    trades.append(orderinfo)
                    status='close'
                if status=='close':
                    if (row['openlogic'] & row['holdlogic']):
                        status = 'buying'
                        orderday = 0
                else:
                    revbase = row['refer']
                    revenue = ((revbase / orderinfo['buybid']) * (1 - self.sellfee) - (1 + self.interest * holdday)) / (self.margin + self.buyfee)
                    guarantee = (revbase / orderinfo['buybid']) * (1 - self.sellfee) - self.interest * holdday + self.margin
                    if guarantee < self.guarantee:
                        status = 'mco'
                        orderday = 0
                    else:
                        hold = row['holdlogic']
                        for logic in self.extralogics:
                            hold = hold & logic.calc(revenue)
                        if hold:
                            status = 'open'
            else:
                raise Exception("Illegal status!")
        if status!='close':
            orderinfo['status']=status
            trades.append(orderinfo)
        return trades

class ShortSell:
    def __init__(self,buyfee,sellfee,interest,margin,guarantee,stopprofit=None,stoploss=None):
        self.buyfee=buyfee
        self.sellfee=sellfee
        self.interest=interest/360
        self.margin=margin
        self.guarantee=guarantee
        self.extralogics = []
        if stopprofit:
            extralogic=stopprofit_logic(stopprofit=stopprofit)
            self.extralogics.append(extralogic)
        if stoploss:
            extralogic=stoploss_logic(stoploss=stoploss)
            self.extralogics.append(extralogic)

    def statistics(self,data):
        if (set(['orderbuy','ordersell','closeout','refer','high','low','openlogic','holdlogic'])<=set(data.columns))==False:
            print('缺少必要数据列',set(['orderbuy','ordersell','closeout','refer','high','low','openlogic','holdlogic']).difference(set(data.columns)))
            return None
        status='close'
        trades=[]
        for inx,row in data.iterrows():
            if status=='close':
                if (row['openlogic']&row['holdlogic']):
                    status='selling'
                    orderday=0
            elif status=='selling':
                orderday+=1
                if (row['ordersell']<=row['high']):
                    status='open'
                    holdday=0
                    orderinfo={'sellinx':inx,'sellday':orderday,'sellbid':max(row['ordersell'],row['low'])}
                if status=='open':
                    revbase=row['refer']
                    revenue=(1-self.sellfee-(revbase/orderinfo['sellbid'])*(1+self.buyfee)-self.interest*holdday)/self.margin
                    guarantee=(1-self.sellfee+self.margin)/((revbase/orderinfo['sellbid'])*(1+self.buyfee)+self.interest*holdday)
                    if guarantee<self.guarantee:
                        status='mco'
                        orderday=0
                    else:
                        hold = row['holdlogic']
                        for logic in self.extralogics:
                            hold = hold & logic.calc(revenue)
                        if hold==False:
                            status = 'buying'
                            orderday = 0
                else:
                    if (row['openlogic']&row['holdlogic'])==False:
                        status='close'
            elif status=='open':
                holdday+=1
                revbase = row['refer']
                revenue = (1 - self.sellfee - (revbase / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * holdday) / self.margin
                guarantee = (1 - self.sellfee + self.margin) / ((revbase / orderinfo['sellbid']) * (1 + self.buyfee) + self.interest * holdday)
                if guarantee < self.guarantee:
                    status = 'mco'
                    orderday = 0
                else:
                    hold = row['holdlogic']
                    for logic in self.extralogics:
                        hold = hold & logic.calc(revenue)
                    if hold == False:
                        status = 'buying'
                        orderday = 0
            elif status=='mco':
                holdday += 1
                orderday+=1
                if row['closeout']>=row['low']:
                    orderinfo['buyinx']=inx
                    orderinfo['buyday']=orderday
                    orderinfo['buybid']=min(row['closeout'],row['high'])
                    orderinfo['revenue']=(1 - self.sellfee - (orderinfo['buybid'] / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * holdday) / self.margin
                    orderinfo['guarantee']=(1 - self.sellfee + self.margin) / ((orderinfo['buybid'] / orderinfo['sellbid']) * (1 + self.buyfee) + self.interest * holdday)
                    orderinfo['cost']=(self.sellfee+(orderinfo['buybid'] / orderinfo['sellbid'])*self.buyfee+self.interest * holdday)/self.margin
                    orderinfo['closetype']='margin-closeout'
                    highpiece=data['high'][orderinfo['sellinx']:orderinfo['buyinx']]
                    lowpiece = data['low'][orderinfo['sellinx']:orderinfo['buyinx']]
                    bppiece=data['orderbuy'][orderinfo['sellinx']:orderinfo['buyinx']]
                    bppiece=(highpiece>=bppiece)*bppiece+(highpiece<bppiece)*highpiece
                    bppiece=(lowpiece<=bppiece)*bppiece+(lowpiece>bppiece)*lowpiece
                    bpdays=pd.Series(range(len(bppiece)),index=bppiece.index)
                    revenueseries=(1 - self.sellfee - (bppiece / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * bpdays) / self.margin
                    orderinfo['revmax']=revenueseries.max()
                    orderinfo['revmin']=revenueseries.min()
                    trades.append(orderinfo)
                    status='close'
            elif status=='buying':
                holdday += 1
                orderday+=1
                if row['orderbuy']>=row['low']:
                    orderinfo['buyinx']=inx
                    orderinfo['buyday']=orderday
                    orderinfo['buybid']=min(row['orderbuy'],row['high'])
                    orderinfo['revenue']=(1 - self.sellfee - (orderinfo['buybid'] / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * holdday) / self.margin
                    orderinfo['guarantee']=(1 - self.sellfee + self.margin) / ((orderinfo['buybid'] / orderinfo['sellbid']) * (1 + self.buyfee) + self.interest * holdday)
                    orderinfo['cost']=(self.sellfee+(orderinfo['buybid'] / orderinfo['sellbid'])*self.buyfee+self.interest * holdday)/self.margin
                    orderinfo['closetype']='buy-close'
                    highpiece=data['high'][orderinfo['sellinx']:orderinfo['buyinx']]
                    lowpiece = data['low'][orderinfo['sellinx']:orderinfo['buyinx']]
                    bppiece=data['orderbuy'][orderinfo['sellinx']:orderinfo['buyinx']]
                    bppiece=(highpiece>=bppiece)*bppiece+(highpiece<bppiece)*highpiece
                    bppiece=(lowpiece<=bppiece)*bppiece+(lowpiece>bppiece)*lowpiece
                    bpdays=pd.Series(range(len(bppiece)),index=bppiece.index)
                    revenueseries=(1 - self.sellfee - (bppiece / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * bpdays) / self.margin
                    orderinfo['revmax']=revenueseries.max()
                    orderinfo['revmin']=revenueseries.min()
                    trades.append(orderinfo)
                    status='close'
                if status=='close':
                    if (row['openlogic'] & row['holdlogic']):
                        status = 'selling'
                        orderday = 0
                else:
                    revbase = row['refer']
                    revenue = (1 - self.sellfee - (revbase / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * holdday) / self.margin
                    guarantee = (1 - self.sellfee + self.margin) / ((revbase / orderinfo['sellbid']) * (1 + self.buyfee) + self.interest * holdday)
                    if guarantee < self.guarantee:
                        status = 'mco'
                        orderday = 0
                    else:
                        hold = row['holdlogic']
                        for logic in self.extralogics:
                            hold = hold & logic.calc(revenue)
                        if hold:
                            status='open'
            else:
                raise Exception("Illegal status!")
        if status!='close':
            orderinfo['status']=status
            trades.append(orderinfo)
        return trades

    def statistics_with_callauction(self,data):
        if (set(['conbuy','consell','closeout','callbuy','callsell','refer','high','low','openlogic','holdlogic'])<=set(data.columns))==False:
            print('缺少必要数据列',set(['conbuy','consell','closeout','callbuy','callsell','refer','high','low','openlogic','holdlogic']).difference(set(data.columns)))
            return None
        status='close'
        trades=[]
        for inx,row in data.iterrows():
            if status=='close':
                if (row['openlogic']&row['holdlogic']):
                    status='selling'
                    orderday=0
            elif status=='selling':
                orderday+=1
                if (row['callsell']<=row['open']):
                    status='open'
                    holdday=0
                    orderinfo={'sellinx':inx,'sellday':orderday,'sellbid':row['open']}
                elif (row['consell']<=row['high']):
                    status='open'
                    holdday=0
                    orderinfo={'sellinx':inx,'sellday':orderday,'sellbid':max(row['consell'],row['low'])}
                if status=='open':
                    revbase=row['refer']
                    revenue=(1-self.sellfee-(revbase/orderinfo['sellbid'])*(1+self.buyfee)-self.interest*holdday)/self.margin
                    guarantee=(1-self.sellfee+self.margin)/((revbase/orderinfo['sellbid'])*(1+self.buyfee)+self.interest*holdday)
                    if guarantee<self.guarantee:
                        status='mco'
                        orderday=0
                    else:
                        hold = row['holdlogic']
                        for logic in self.extralogics:
                            hold = hold & logic.calc(revenue)
                        if hold==False:
                            status = 'buying'
                            orderday = 0
                else:
                    if (row['openlogic']&row['holdlogic'])==False:
                        status='close'
            elif status=='open':
                holdday+=1
                revbase = row['refer']
                revenue = (1 - self.sellfee - (revbase / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * holdday) / self.margin
                guarantee = (1 - self.sellfee + self.margin) / ((revbase / orderinfo['sellbid']) * (1 + self.buyfee) + self.interest * holdday)
                if guarantee < self.guarantee:
                    status = 'mco'
                    orderday = 0
                else:
                    hold = row['holdlogic']
                    for logic in self.extralogics:
                        hold = hold & logic.calc(revenue)
                    if hold == False:
                        status = 'buying'
                        orderday = 0
            elif status=='mco':
                holdday += 1
                orderday+=1
                if row['closeout']>=row['open']:
                    orderinfo['buyinx']=inx
                    orderinfo['buyday']=orderday
                    orderinfo['buybid']=row['open']
                    orderinfo['revenue']=(1 - self.sellfee - (orderinfo['buybid'] / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * holdday) / self.margin
                    orderinfo['guarantee']=(1 - self.sellfee + self.margin) / ((orderinfo['buybid'] / orderinfo['sellbid']) * (1 + self.buyfee) + self.interest * holdday)
                    orderinfo['cost']=(self.sellfee+(orderinfo['buybid'] / orderinfo['sellbid'])*self.buyfee+self.interest * holdday)/self.margin
                    orderinfo['closetype']='margin-closeout'
                    highpiece=data['high'][orderinfo['sellinx']:orderinfo['buyinx']]
                    lowpiece = data['low'][orderinfo['sellinx']:orderinfo['buyinx']]
                    bppiece=data['refer'].shift()[orderinfo['sellinx']:orderinfo['buyinx']]
                    bppiece=(highpiece>=bppiece)*bppiece+(highpiece<bppiece)*highpiece
                    bppiece=(lowpiece<=bppiece)*bppiece+(lowpiece>bppiece)*lowpiece
                    bpdays=pd.Series(range(len(bppiece)),index=bppiece.index)
                    revenueseries=(1 - self.sellfee - (bppiece / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * bpdays) / self.margin
                    orderinfo['revmax']=revenueseries.max()
                    orderinfo['revmin']=revenueseries.min()
                    trades.append(orderinfo)
                    status='close'
                elif row['closeout']>=row['low']:
                    orderinfo['buyinx']=inx
                    orderinfo['buyday']=orderday
                    orderinfo['buybid']=min(row['closeout'],row['high'])
                    orderinfo['revenue']=(1 - self.sellfee - (orderinfo['buybid'] / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * holdday) / self.margin
                    orderinfo['guarantee']=(1 - self.sellfee + self.margin) / ((orderinfo['buybid'] / orderinfo['sellbid']) * (1 + self.buyfee) + self.interest * holdday)
                    orderinfo['cost']=(self.sellfee+(orderinfo['buybid'] / orderinfo['sellbid'])*self.buyfee+self.interest * holdday)/self.margin
                    orderinfo['closetype']='margin-closeout'
                    highpiece=data['high'][orderinfo['sellinx']:orderinfo['buyinx']]
                    lowpiece = data['low'][orderinfo['sellinx']:orderinfo['buyinx']]
                    bppiece=data['refer'].shift()[orderinfo['sellinx']:orderinfo['buyinx']]
                    bppiece=(highpiece>=bppiece)*bppiece+(highpiece<bppiece)*highpiece
                    bppiece=(lowpiece<=bppiece)*bppiece+(lowpiece>bppiece)*lowpiece
                    bpdays=pd.Series(range(len(bppiece)),index=bppiece.index)
                    revenueseries=(1 - self.sellfee - (bppiece / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * bpdays) / self.margin
                    orderinfo['revmax']=revenueseries.max()
                    orderinfo['revmin']=revenueseries.min()
                    trades.append(orderinfo)
                    status='close'
            elif status=='buying':
                holdday += 1
                orderday+=1
                if row['callbuy']>=row['open']:
                    orderinfo['buyinx']=inx
                    orderinfo['buyday']=orderday
                    orderinfo['buybid']=row['open']
                    orderinfo['revenue']=(1 - self.sellfee - (orderinfo['buybid'] / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * holdday) / self.margin
                    orderinfo['guarantee']=(1 - self.sellfee + self.margin) / ((orderinfo['buybid'] / orderinfo['sellbid']) * (1 + self.buyfee) + self.interest * holdday)
                    orderinfo['cost']=(self.sellfee+(orderinfo['buybid'] / orderinfo['sellbid'])*self.buyfee+self.interest * holdday)/self.margin
                    orderinfo['closetype']='buy-close'
                    highpiece=data['high'][orderinfo['sellinx']:orderinfo['buyinx']]
                    lowpiece = data['low'][orderinfo['sellinx']:orderinfo['buyinx']]
                    bppiece=data['refer'].shift()[orderinfo['sellinx']:orderinfo['buyinx']]
                    bppiece=(highpiece>=bppiece)*bppiece+(highpiece<bppiece)*highpiece
                    bppiece=(lowpiece<=bppiece)*bppiece+(lowpiece>bppiece)*lowpiece
                    bpdays=pd.Series(range(len(bppiece)),index=bppiece.index)
                    revenueseries=(1 - self.sellfee - (bppiece / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * bpdays) / self.margin
                    orderinfo['revmax']=revenueseries.max()
                    orderinfo['revmin']=revenueseries.min()
                    trades.append(orderinfo)
                    status='close'
                elif row['conbuy']>=row['low']:
                    orderinfo['buyinx']=inx
                    orderinfo['buyday']=orderday
                    orderinfo['buybid']=min(row['conbuy'],row['high'])
                    orderinfo['revenue']=(1 - self.sellfee - (orderinfo['buybid'] / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * holdday) / self.margin
                    orderinfo['guarantee']=(1 - self.sellfee + self.margin) / ((orderinfo['buybid'] / orderinfo['sellbid']) * (1 + self.buyfee) + self.interest * holdday)
                    orderinfo['cost']=(self.sellfee+(orderinfo['buybid'] / orderinfo['sellbid'])*self.buyfee+self.interest * holdday)/self.margin
                    orderinfo['closetype']='buy-close'
                    highpiece=data['high'][orderinfo['sellinx']:orderinfo['buyinx']]
                    lowpiece = data['low'][orderinfo['sellinx']:orderinfo['buyinx']]
                    bppiece=data['refer'].shift()[orderinfo['sellinx']:orderinfo['buyinx']]
                    bppiece=(highpiece>=bppiece)*bppiece+(highpiece<bppiece)*highpiece
                    bppiece=(lowpiece<=bppiece)*bppiece+(lowpiece>bppiece)*lowpiece
                    bpdays=pd.Series(range(len(bppiece)),index=bppiece.index)
                    revenueseries=(1 - self.sellfee - (bppiece / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * bpdays) / self.margin
                    orderinfo['revmax']=revenueseries.max()
                    orderinfo['revmin']=revenueseries.min()
                    trades.append(orderinfo)
                    status='close'
                if status=='close':
                    if (row['openlogic'] & row['holdlogic']):
                        status = 'selling'
                        orderday = 0
                else:
                    revbase = row['refer']
                    revenue = (1 - self.sellfee - (revbase / orderinfo['sellbid']) * (1 + self.buyfee) - self.interest * holdday) / self.margin
                    guarantee = (1 - self.sellfee + self.margin) / ((revbase / orderinfo['sellbid']) * (1 + self.buyfee) + self.interest * holdday)
                    if guarantee < self.guarantee:
                        status = 'mco'
                        orderday = 0
                    else:
                        hold = row['holdlogic']
                        for logic in self.extralogics:
                            hold = hold & logic.calc(revenue)
                        if hold:
                            status='open'
            else:
                raise Exception("Illegal status!")
        if status!='close':
            orderinfo['status']=status
            trades.append(orderinfo)
        return trades

class stopprofit_logic:
    def __init__(self,stopprofit):
        self.stopprofit=stopprofit

    def calc(self,revenue):
        return revenue<self.stopprofit

class stoploss_logic:
    def __init__(self,stoploss):
        self.stoploss=stoploss

    def calc(self,revenue):
        return revenue>self.stoploss







# 著作者：叶凯文
# Author: YE KAIWEN
# 版权归著作者叶凯文所有
# Copyright (c) 2022 YE KAIWEN. All Rights Reserved.

import numpy as np
import pandas as pd
from .Models import common_model
from .StockModels import company_model,AFREQS
from .ModelSets import common_modelset
from .StockModelSets import company_modelset

class ConnectionMethod1:
    def __init__(self,dtpow=2,fmpow=2,dtsopen=1/2,fmsopen=1/2):
        self.dtpow=dtpow
        self.fmpow=fmpow
        self.dtsopen=dtsopen
        self.fmsopen=fmsopen

    def seriesmethod(self,series1,series2):
        distance = pow(pow(series1-series2,self.dtpow).sum(),self.dtsopen)
        formator = pow(pow(series1+series2,self.fmpow).sum(),self.fmsopen)
        return distance / formator

    def elementmethod(self,element1,element2):
        distance = pow(pow(element1-element2,self.dtpow),self.dtsopen)
        formator = pow(pow(element1+element2,self.fmpow),self.fmsopen)
        return distance / formator

class ConnectionMethod2:
    def __init__(self,dtpow=2,fmpow=2,dtsopen=1/2,fmsopen=1/2):
        self.dtpow=dtpow
        self.fmpow=fmpow
        self.dtsopen=dtsopen
        self.fmsopen=fmsopen

    def seriesmethod(self,series1,series2):
        distance = pow(pow(series1-series2,self.dtpow).sum(),self.dtsopen)
        formator = pow((pow(series1,self.fmpow)+pow(series2,self.fmpow)).sum(),self.fmsopen)
        return distance / formator

    def elementmethod(self,element1,element2):
        distance = pow(pow(element1-element2,self.dtpow),self.dtsopen)
        formator = pow((pow(element1,self.fmpow)+pow(element2,self.fmpow)),self.fmsopen)
        return distance / formator

class ConnectionMethod3:
    def __init__(self,coef1,coef2):
        self.coef1=coef1
        self.coef2=coef2

    def seriesmethod(self,series1,series2):
        return self.coef1*np.exp(pow(series1-series2,2).sum()/(2*self.coef2))

    def elementmethod(self,element1,element2):
        return self.coef1 * np.exp(pow(element1 - element2, 2) / (2 * self.coef2))

class ConnectionMethod4:
    def seriesmethod(self,series1,series2):
        simularity=(series1*series2/(series1+series2)).sum()
        if simularity==0:
            return np.inf
        else:
            return 1/simularity

    def elementmethod(self,element1,element2):
        simularity=(element1*element2/(element1+element2))
        if simularity == 0:
            return np.inf
        else:
            return 1 / simularity

class ConnectionMethod5:
    def seriesmethod(self,series1,series2):
        simularity=(series1*series2).sum()/np.sqrt(pow(series1,2).sum()*pow(series2,2).sum())
        if simularity == 0:
            return np.inf
        else:
            return 1 / simularity

    def elementmethod(self,element1,element2):
        simularity=(element1*element2)/np.sqrt(pow(element1,2)*pow(element2,2))
        if simularity == 0:
            return np.inf
        else:
            return 1 / simularity

class CommonResonanceModel:
    def __init__(self,models,connection):
        self._models=models
        self._connection=connection
        self.__getfreqparams()

    def __getfreqparams(self):
        if isinstance(self._models[0], common_model):
            self._freq=self._models[0]._mparams['freq']
            self._AP=len(pd.date_range(start='2022-01-01',end='2022-12-31',freq=self._freq))
        elif isinstance(self._models[0], company_model):
            if (self._models[0]._mparams['ttm']):
                self._freq=AFREQS[self._models[0]._mparams['reportdate'][-4:]]
                self._AP=1
            else:
                self._freq='Q-DEC'
                self._AP = 4
        elif isinstance(self._models[0], StockSetDimensionModel):
            if (self._models[0]._mparams['ttm']):
                self._freq=AFREQS[self._models[0]._mparams['reportdate'][-4:]]
                self._AP=1
            else:
                self._freq='Q-DEC'
                self._AP = 4
        elif isinstance(self._models[0], ModelSetDimensionModel):
            self._freq=self._models[0]._mparams['freq']
            self._AP=len(pd.date_range(start='2022-01-01',end='2022-12-31',freq=self._freq))
        else:
            raise Exception("模型类型错误")
        reportlist=[]
        for model in self._models:
            if isinstance(model,common_model):
                if (model._mparams['freq']==self._freq):
                    reportlist.append(model._data.index[-1])
                else:
                    raise Exception("模型类型错误")
            elif isinstance(model,company_model):
                if (model._mparams['ttm']):
                    if (AFREQS[model._mparams['reportdate'][-4:]]==self._freq):
                        reportlist.append(pd.to_datetime(model._mparams['reportdate']))
                    else:
                        raise Exception("模型类型错误")
                else:
                    reportlist.append(pd.to_datetime(model._mparams['reportdate']))
            elif isinstance(model,StockSetDimensionModel):
                if (model._mparams['ttm']):
                    if (AFREQS[model._mparams['reportdate'][-4:]]==self._freq):
                        reportlist.append(pd.to_datetime(model._mparams['reportdate']))
                    else:
                        raise Exception("模型类型错误")
                else:
                    reportlist.append(pd.to_datetime(model._mparams['reportdate']))
            elif isinstance(model,ModelSetDimensionModel):
                if (model._mparams['freq']==self._freq):
                    reportlist.append(model.data().index[-1])
                else:
                    raise Exception("模型类型错误")
            else:
                raise Exception("模型类型错误")
        reportmax=np.array(reportlist).max()
        periodgaps=[]
        for report in reportlist:
            periodgaps.append(len(pd.date_range(start=report,end=reportmax,freq=self._freq))-1)
        self._periodgaps=periodgaps

    def forecast(self,t,droplow=0):
        fores=[]
        mnum=len(self._models)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        for i in range(mnum):
            foreseries=self._models[i].forecast(t+self._periodgaps[i])
            fores.append(foreseries.iloc[self._periodgaps[i]:])
        if len(fores)==0:
            return None
        elif len(fores)==1:
            return fores[0]
        fores=pd.DataFrame(fores).transpose()
        resonances=[]
        for i in range(mnum):
            cw=0
            for j in range(mnum):
                if i==j:
                    continue
                cw+=self._connection.seriesmethod(fores[i],fores[j])
            if cw==0:
                return fores[0]
            resonances.append(1/cw)
        resonances=pd.Series(resonances)
        for m in range(droplow):
            dinx=resonances.idxmin()
            del resonances[dinx]
            del fores[dinx]
        return (fores*resonances).sum(1)/(resonances.sum())

    def pv(self,r,droplow=0):
        pvs=[]
        mnum = len(self._models)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        sr=pow(1+r,1/self._AP)-1
        for i in range(mnum):
            if self._models[i]._mparams['seasonal']:
                pvs.append(self._models[i].pv(r)*pow(1+r,self._periodgaps[i]/self._AP))
            else:
                pvs.append(self._models[i].pv(sr) * pow(1+r, self._periodgaps[i] / self._AP))
        if len(pvs)==0:
            return None
        elif len(pvs)==1:
            return pvs[0]
        pvs=pd.Series(pvs)
        resonances=[]
        for i in range(mnum):
            cw=0
            for j in range(mnum):
                if i==j:
                    continue
                cw+=self._connection.elementmethod(pvs[i],pvs[j])
            if cw==0:
                return pvs[0]
            resonances.append(1/cw)
        resonances=pd.Series(resonances)
        for m in range(droplow):
            dinx=resonances.idxmin()
            del resonances[dinx]
            del pvs[dinx]
        return (pvs*resonances).sum()/(resonances.sum())

    def irr(self,price,tradedate,droplow=0):
        rates=[]
        mnum = len(self._models)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        for i in range(mnum):
            irr = self._models[i].irr(price, tradedate)
            if type(irr) == float:
                if self._models[i]._mparams['seasonal']:
                    rates.append(1+irr)
                else:
                    rates.append(pow(1 + irr,self._AP))
        if len(rates)==0:
            return None
        elif len(rates)==1:
            return rates[0]-1
        rates=np.log(pd.Series(rates))
        resonances=[]
        for i in range(len(rates)):
            cw=0
            for j in range(len(rates)):
                if i==j:
                    continue
                cw+=self._connection.elementmethod(rates[i],rates[j])
            if cw==0:
                return np.exp(rates[0])-1
            resonances.append(1/cw)
        resonances=pd.Series(resonances)
        for m in range(droplow-(mnum-len(rates))):
            dinx=resonances.idxmin()
            del resonances[dinx]
            del rates[dinx]
        if len(rates)>0:
            return np.exp((rates * resonances).sum() / (resonances.sum()))-1
        else:
            return None

    def batchirr(self,priceseries,droplow=0):
        batchrates=[]
        mnum = len(self._models)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        for i in range(mnum):
            if self._models[i]._mparams['seasonal']:
                batchrates.append(1+self._models[i].batchirr(priceseries)['IRRs'])
            else:
                batchrates.append(pow(1+self._models[i].batchirr(priceseries)['IRRs'],self._AP))
        batchrates=np.log(pd.DataFrame(batchrates).transpose())
        irrs=pd.Series(index=batchrates.index)
        for index,row in batchrates.iterrows():
            rates=row.copy(deep=True).dropna()
            if len(rates) == 0:
                irrs[index]=np.nan
                continue
            elif len(rates) == 1:
                irrs[index]=np.exp(rates.values[0])-1
                continue
            rates=rates.reset_index(drop=True)
            resonances = []
            homo=False
            for i in range(len(rates)):
                cw = 0
                for j in range(len(rates)):
                    if i == j:
                        continue
                    cw += self._connection.elementmethod(rates[i], rates[j])
                if cw == 0:
                    homo = True
                    break
                resonances.append(1 / cw)
            if homo:
                resonances = pd.Series(np.ones(len(rates)))
            else:
                resonances = pd.Series(resonances)
            for m in range(droplow-(mnum-len(rates))):
                dinx = resonances.idxmin()
                del resonances[dinx]
                del rates[dinx]
            if len(rates)>0:
                irrs[index] = np.exp((rates * resonances).sum() / (resonances.sum()))-1
            else:
                irrs[index]=np.nan
        return irrs

    def achieverate(self,price,r,tradedate,droplow=0):
        acrs = []
        mnum = len(self._models)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        sr = pow(1 + r, 1 / self._AP) - 1
        for i in range(mnum):
            if self._models[i]._mparams['seasonal']:
                acrs.append(self._models[i].achieverate(price,r,tradedate))
            else:
                acrs.append(self._models[i].achieverate(price,sr,tradedate))
        if len(acrs) == 0:
            return None
        elif len(acrs) == 1:
            return acrs[0]
        acrs = pd.Series(acrs)
        resonances = []
        for i in range(mnum):
            cw = 0
            for j in range(mnum):
                if i == j:
                    continue
                cw += self._connection.elementmethod(acrs[i], acrs[j])
            if cw == 0:
                return acrs[0]
            resonances.append(1 / cw)
        resonances = pd.Series(resonances)
        for m in range(droplow):
            dinx=resonances.idxmin()
            del resonances[dinx]
            del acrs[dinx]
        return (acrs*resonances).sum()/(resonances.sum())

    def batchachieverate(self,priceseries,r,droplow=0):
        batchacrs=[]
        mnum = len(self._models)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        sr = pow(1 + r, 1 / self._AP) - 1
        for i in range(mnum):
            if self._models[i]._mparams['seasonal']:
                batchacrs.append(self._models[i].batchachieverate(priceseries,r))
            else:
                batchacrs.append(self._models[i].batchachieverate(priceseries,sr))
        batchacrs=pd.DataFrame(batchacrs).transpose()
        acrs=pd.Series(index=batchacrs.index)
        for index,row in batchacrs.iterrows():
            rates=row.copy(deep=True)
            if len(rates) == 0:
                acrs[index]=np.nan
                continue
            elif len(rates) == 1:
                acrs[index]=rates.values[0]
                continue
            resonances = []
            homo = False
            for i in range(mnum):
                cw = 0
                for j in range(mnum):
                    if i == j:
                        continue
                    cw += self._connection.elementmethod(rates[i], rates[j])
                if cw == 0:
                    homo = True
                    break
                resonances.append(1 / cw)
            if homo:
                resonances = pd.Series(np.ones(len(rates)))
            else:
                resonances = pd.Series(resonances)
            for m in range(droplow):
                dinx = resonances.idxmin()
                del resonances[dinx]
                del rates[dinx]
            acrs[index] = (rates*resonances).sum()/(resonances.sum())
        return acrs

class CompanyResonanceModel(CommonResonanceModel):
    def __init__(self,models,connection):
        self._models=models
        self._connection=connection
        self.__getfreqparams()

    def __getfreqparams(self):
        if isinstance(self._models[0], company_model):
            if (self._models[0]._mparams['ttm']):
                self._freq=AFREQS[self._models[0]._mparams['reportdate'][-4:]]
                self._AP=1
            else:
                self._freq='Q-DEC'
                self._AP = 4
        elif isinstance(self._models[0], StockSetDimensionModel):
            if (self._models[0]._mparams['ttm']):
                self._freq=AFREQS[self._models[0]._mparams['reportdate'][-4:]]
                self._AP=1
            else:
                self._freq='Q-DEC'
                self._AP = 4
        else:
            raise Exception("模型类型错误")
        reportlist=[]
        for model in self._models:
            if isinstance(model,company_model):
                if (model._mparams['ttm']):
                    if (AFREQS[model._mparams['reportdate'][-4:]]==self._freq):
                        reportlist.append(pd.to_datetime(model._mparams['reportdate']))
                    else:
                        raise Exception("模型类型错误")
                else:
                    reportlist.append(pd.to_datetime(model._mparams['reportdate']))
            elif isinstance(model,StockSetDimensionModel):
                if (model._mparams['ttm']):
                    if (AFREQS[model._mparams['reportdate'][-4:]]==self._freq):
                        reportlist.append(pd.to_datetime(model._mparams['reportdate']))
                    else:
                        raise Exception("模型类型错误")
                else:
                    reportlist.append(pd.to_datetime(model._mparams['reportdate']))
            else:
                raise Exception("模型类型错误")
        reportmax=np.array(reportlist).max()
        periodgaps=[]
        for report in reportlist:
            periodgaps.append(len(pd.date_range(start=report,end=reportmax,freq=self._freq))-1)
        self._periodgaps=periodgaps

    def irr_trade(self,pricemode,tradedate,droplow=0):
        rates=[]
        mnum = len(self._models)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        for i in range(mnum):
            irr=self._models[i].irr_trade(pricemode,tradedate)
            if type(irr)==float:
                if self._models[i]._mparams['seasonal']:
                    rates.append(1+irr)
                else:
                    rates.append(pow(1 + irr,self._AP))
        if len(rates)==0:
            return None
        elif len(rates)==1:
            return rates[0]-1
        rates=np.log(pd.Series(rates))
        resonances=[]
        for i in range(len(rates)):
            cw=0
            for j in range(len(rates)):
                if i==j:
                    continue
                cw+=self._connection.elementmethod(rates[i],rates[j])
            if cw==0:
                return np.exp(rates[0])-1
            resonances.append(1/cw)
        resonances=pd.Series(resonances)
        for m in range(droplow-(mnum-len(rates))):
            dinx=resonances.idxmin()
            del resonances[dinx]
            del rates[dinx]
        if len(rates)>0:
            return np.exp((rates * resonances).sum() / (resonances.sum()))-1
        else:
            return None

    def batchirr_trade(self,pricemode,startdate,enddate,droplow=0):
        batchrates=[]
        mnum = len(self._models)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        for i in range(mnum):
            if self._models[i]._mparams['seasonal']:
                batchrates.append(1+self._models[i].batchirr_trade(pricemode,startdate,enddate)['IRRs'])
            else:
                batchrates.append(pow(1+self._models[i].batchirr_trade(pricemode,startdate,enddate)['IRRs'],self._AP))
        batchrates=np.log(pd.DataFrame(batchrates).transpose())
        irrs=pd.Series(index=batchrates.index)
        for index,row in batchrates.iterrows():
            rates=row.copy(deep=True).dropna()
            if len(rates) == 0:
                irrs[index] = np.nan
                continue
            elif len(rates) == 1:
                irrs[index] = np.exp(rates.values[0]) - 1
                continue
            rates=rates.reset_index(drop=True)
            resonances = []
            homo = False
            for i in range(len(rates)):
                cw = 0
                for j in range(len(rates)):
                    if i == j:
                        continue
                    cw += self._connection.elementmethod(rates[i], rates[j])
                if cw == 0:
                    homo = True
                    break
                resonances.append(1 / cw)
            if homo:
                resonances = pd.Series(np.ones(len(rates)))
            else:
                resonances = pd.Series(resonances)
            for m in range(droplow-(mnum-len(rates))):
                dinx = resonances.idxmin()
                del resonances[dinx]
                del rates[dinx]
            if len(rates)>0:
                irrs[index] = np.exp((rates * resonances).sum() / (resonances.sum()))-1
            else:
                irrs[index] = np.nan
        return irrs

    def achieverate_trade(self, pricemode, tradedate, r,droplow=0):
        acrs = []
        mnum = len(self._models)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        sr = pow(1 + r, 1 / self._AP) - 1
        for i in range(mnum):
            if self._models[i]._mparams['seasonal']:
                acrs.append(self._models[i].achieverate_trade(pricemode, tradedate, r))
            else:
                acrs.append(self._models[i].achieverate_trade(pricemode, tradedate, sr))
        if len(acrs) == 0:
            return None
        elif len(acrs) == 1:
            return acrs[0]
        acrs = pd.Series(acrs)
        resonances = []
        for i in range(mnum):
            cw = 0
            for j in range(mnum):
                if i == j:
                    continue
                cw += self._connection.elementmethod(acrs[i], acrs[j])
            if cw == 0:
                return acrs[0]
            resonances.append(1 / cw)
        resonances = pd.Series(resonances)
        for m in range(droplow):
            dinx=resonances.idxmin()
            del resonances[dinx]
            del acrs[dinx]
        return (acrs*resonances).sum()/(resonances.sum())

    def batchachieverate_trade(self, pricemode, startdate, enddate, r,droplow=0):
        batchacrs=[]
        mnum = len(self._models)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        sr = pow(1 + r, 1 / self._AP) - 1
        for i in range(mnum):
            if self._models[i]._mparams['seasonal']:
                batchacrs.append(self._models[i].batchachieverate_trade(pricemode, startdate, enddate, r))
            else:
                batchacrs.append(self._models[i].batchachieverate_trade(pricemode, startdate, enddate,sr))
        batchacrs=pd.DataFrame(batchacrs).transpose()
        acrs=pd.Series(index=batchacrs.index)
        for index,row in batchacrs.iterrows():
            rates=row.copy(deep=True)
            if len(rates) == 0:
                acrs[index]=np.nan
                continue
            elif len(rates) == 1:
                acrs[index]=rates.values[0]
                continue
            resonances = []
            homo = False
            for i in range(mnum):
                cw = 0
                for j in range(mnum):
                    if i == j:
                        continue
                    cw += self._connection.elementmethod(rates[i], rates[j])
                if cw == 0:
                    homo = True
                    break
                resonances.append(1 / cw)
            if homo:
                resonances = pd.Series(np.ones(len(rates)))
            else:
                resonances = pd.Series(resonances)
            for m in range(droplow):
                dinx = resonances.idxmin()
                del resonances[dinx]
                del rates[dinx]
            acrs[index] = (rates*resonances).sum()/(resonances.sum())
        return acrs

class ModelSetDimensionModel:
    def __init__(self,modelset,dimension):
        self._modelset=modelset
        self._dimension=dimension
        self._mparams=modelset._mparams

    def model(self):
        return self._modelset.modelset()[self._dimension]

    def fit(self):
        return self._modelset.fit(self._dimension)

    def forecast(self,t):
        return self._modelset.forecast(self._dimension,t)

    def pv(self,r):
        return self._modelset.pv(self._dimension,r)

    def irr(self,price,tradedate):
        return self._modelset.irr(self._dimension,price,tradedate)

    def batchirr(self,priceseries):
        return self._modelset.batchirr(self._dimension,priceseries)

    def achieverate(self,price,r,tradedate):
        return self._modelset.achieverate(self._dimension,price,r,tradedate)

    def batchachieverate(self,priceseries,r):
        return self._modelset.batchachieverate(self._dimension,priceseries,r)

    def data(self):
        return self._modelset._data[self._dimension]

class StockSetDimensionModel(ModelSetDimensionModel):
    def irr_trade(self, pricemode, tradedate):
        return self._modelset.irr_trade(self._dimension, pricemode, tradedate)

    def batchirr_trade(self, pricemode, startdate, enddate):
        return self._modelset.batchirr_trade(self._dimension, pricemode, startdate, enddate)

    def achieverate_trade(self,pricemode, tradedate, r):
        return self._modelset.achieverate_trade(self._dimension,pricemode, tradedate, r)

    def batchachieverate_trade(self, pricemode, startdate, enddate, r):
        return self._modelset.batchachieverate_trade(self._dimension, pricemode, startdate, enddate, r)

    def incdata(self):
        return self._modelset.incdata(self._dimension)

class CommonResonanceModelSet:
    def __init__(self,modelsets,connection):
        self._modelsets=modelsets
        self._connection=connection
        self.__getfreqparams()

    def __getfreqparams(self):
        if isinstance(self._modelsets[0], common_modelset):
            self._freq = self._modelsets[0]._mparams['freq']
            self._AP = len(pd.date_range(start='2022-01-01', end='2022-12-31', freq=self._freq))
        elif isinstance(self._modelsets[0], company_modelset):
            if (self._modelsets[0]._mparams['ttm']):
                self._freq = AFREQS[self._modelsets[0]._mparams['reportdate'][-4:]]
                self._AP = 1
            else:
                self._freq = 'Q-DEC'
                self._AP = 4
        else:
            raise Exception("模型类型错误")
        reportlist = []
        for modelset in self._modelsets:
            if isinstance(modelset, common_modelset):
                if (modelset._mparams['freq'] == self._freq):
                    reportlist.append(modelset._data.index[-1])
                else:
                    raise Exception("模型类型错误")
            elif isinstance(modelset, company_modelset):
                if (modelset._mparams['ttm']) :
                    if (AFREQS[modelset._mparams['reportdate'][-4:]] == self._freq):
                        reportlist.append(pd.to_datetime(modelset._mparams['reportdate']))
                    else:
                        raise Exception("模型类型错误")
                else:
                    reportlist.append(pd.to_datetime(modelset._mparams['reportdate']))
            else:
                raise Exception("模型类型错误")
        reportmax = np.array(reportlist).max()
        periodgaps = []
        for report in reportlist:
            periodgaps.append(len(pd.date_range(start=report, end=reportmax, freq=self._freq)) - 1)
        self._periodgaps = periodgaps

    def forecast(self,dimension,t,droplow=0):
        fores=[]
        mnum=len(self._modelsets)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        for i in range(mnum):
            foreseries=self._modelsets[i].forecast(dimension,t+self._periodgaps[i])
            fores.append(foreseries.iloc[self._periodgaps[i]:])
        if len(fores)==0:
            return None
        elif len(fores)==1:
            return fores[0]
        fores=pd.DataFrame(fores).transpose()
        resonances=[]
        for i in range(mnum):
            cw=0
            for j in range(mnum):
                if i==j:
                    continue
                cw+=self._connection.seriesmethod(fores[i],fores[j])
            if cw==0:
                return fores[0]
            resonances.append(1/cw)
        resonances=pd.Series(resonances)
        for m in range(droplow):
            dinx=resonances.idxmin()
            del resonances[dinx]
            del fores[dinx]
        return (fores*resonances).sum(1)/(resonances.sum())

    def pv(self,dimension,r,droplow=0):
        pvs=[]
        mnum = len(self._modelsets)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        sr=pow(1+r,1/self._AP)-1
        for i in range(mnum):
            if self._modelsets[i]._mparams['seasonal']:
                pvs.append(self._modelsets[i].pv(dimension,r)*pow(1+r,self._periodgaps[i]/self._AP))
            else:
                pvs.append(self._modelsets[i].pv(dimension,sr) * pow(1+r, self._periodgaps[i] / self._AP))
        if len(pvs)==0:
            return None
        elif len(pvs)==1:
            return pvs[0]
        pvs=pd.Series(pvs)
        resonances=[]
        for i in range(mnum):
            cw=0
            for j in range(mnum):
                if i==j:
                    continue
                cw+=self._connection.elementmethod(pvs[i],pvs[j])
            if cw==0:
                return pvs[0]
            resonances.append(1/cw)
        resonances=pd.Series(resonances)
        for m in range(droplow):
            dinx=resonances.idxmin()
            del resonances[dinx]
            del pvs[dinx]
        return (pvs*resonances).sum()/(resonances.sum())

    def irr(self,dimension,price,tradedate,droplow=0):
        rates=[]
        mnum = len(self._modelsets)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        for i in range(mnum):
            irr=self._modelsets[i].irr(dimension,price,tradedate)
            if type(irr)==float:
                if self._modelsets[i]._mparams['seasonal']:
                    rates.append(1+irr)
                else:
                    rates.append(pow(1 + irr,self._AP))
        if len(rates)==0:
            return None
        elif len(rates)==1:
            return rates[0]-1
        rates=np.log(pd.Series(rates))
        resonances=[]
        for i in range(len(rates)):
            cw=0
            for j in range(len(rates)):
                if i==j:
                    continue
                cw+=self._connection.elementmethod(rates[i],rates[j])
            if cw==0:
                return np.exp(rates[0])-1
            resonances.append(1/cw)
        resonances=pd.Series(resonances)
        for m in range(droplow-(mnum-len(rates))):
            dinx=resonances.idxmin()
            del resonances[dinx]
            del rates[dinx]
        if len(rates)>0:
            return np.exp((rates * resonances).sum() / (resonances.sum()))-1
        else:
            return None

    def batchirr(self,dimension,priceseries,droplow=0):
        batchrates=[]
        mnum = len(self._modelsets)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        for i in range(mnum):
            if self._modelsets[i]._mparams['seasonal']:
                batchrates.append(1+self._modelsets[i].batchirr(dimension,priceseries)['IRRs'])
            else:
                batchrates.append(pow(1+self._modelsets[i].batchirr(dimension,priceseries)['IRRs'],self._AP))
        batchrates=np.log(pd.DataFrame(batchrates).transpose())
        irrs=pd.Series(index=batchrates.index)
        for index,row in batchrates.iterrows():
            rates=row.copy(deep=True).dropna()
            if len(rates) == 0:
                irrs[index] = np.nan
                continue
            elif len(rates) == 1:
                irrs[index] = np.exp(rates.values[0]) - 1
                continue
            rates=rates.reset_index(drop=True)
            resonances = []
            homo = False
            for i in range(len(rates)):
                cw = 0
                for j in range(len(rates)):
                    if i == j:
                        continue
                    cw += self._connection.elementmethod(rates[i], rates[j])
                if cw == 0:
                    homo = True
                    break
                resonances.append(1 / cw)
            if homo:
                resonances = pd.Series(np.ones(len(rates)))
            else:
                resonances = pd.Series(resonances)
            for m in range(droplow-(mnum-len(rates))):
                dinx = resonances.idxmin()
                del resonances[dinx]
                del rates[dinx]
            if len(rates)>0:
                irrs[index] = np.exp((rates * resonances).sum() / (resonances.sum()))-1
            else:
                irrs[index] = np.nan
        return irrs

    def achieverate(self,dimension,price,r,tradedate,droplow=0):
        acrs = []
        mnum = len(self._modelsets)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        sr = pow(1 + r, 1 / self._AP) - 1
        for i in range(mnum):
            if self._modelsets[i]._mparams['seasonal']:
                acrs.append(self._modelsets[i].achieverate(dimension,price,r,tradedate))
            else:
                acrs.append(self._modelsets[i].achieverate(dimension,price,sr,tradedate))
        if len(acrs) == 0:
            return None
        elif len(acrs) == 1:
            return acrs[0]
        acrs = pd.Series(acrs)
        resonances = []
        for i in range(mnum):
            cw = 0
            for j in range(mnum):
                if i == j:
                    continue
                cw += self._connection.elementmethod(acrs[i], acrs[j])
            if cw == 0:
                return acrs[0]
            resonances.append(1 / cw)
        resonances = pd.Series(resonances)
        for m in range(droplow):
            dinx=resonances.idxmin()
            del resonances[dinx]
            del acrs[dinx]
        return (acrs*resonances).sum()/(resonances.sum())

    def batchachieverate(self,dimension,priceseries,r,droplow=0):
        batchacrs=[]
        mnum = len(self._modelsets)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        sr = pow(1 + r, 1 / self._AP) - 1
        for i in range(mnum):
            if self._modelsets[i]._mparams['seasonal']:
                batchacrs.append(self._modelsets[i].batchachieverate(dimension,priceseries,r))
            else:
                batchacrs.append(self._modelsets[i].batchachieverate(dimension,priceseries,sr))
        batchacrs=pd.DataFrame(batchacrs).transpose()
        acrs=pd.Series(index=batchacrs.index)
        for index,row in batchacrs.iterrows():
            rates=row.copy(deep=True)
            if len(rates) == 0:
                acrs[index]=np.nan
                continue
            elif len(rates) == 1:
                acrs[index]=rates.values[0]
                continue
            resonances = []
            homo = False
            for i in range(mnum):
                cw = 0
                for j in range(mnum):
                    if i == j:
                        continue
                    cw += self._connection.elementmethod(rates[i], rates[j])
                if cw == 0:
                    homo = True
                    break
                resonances.append(1 / cw)
            if homo:
                resonances = pd.Series(np.ones(len(rates)))
            else:
                resonances = pd.Series(resonances)
            for m in range(droplow):
                dinx = resonances.idxmin()
                del resonances[dinx]
                del rates[dinx]
            acrs[index] = (rates*resonances).sum()/(resonances.sum())
        return acrs

class CompanyResonanceModelSet(CommonResonanceModelSet):
    def __init__(self,modelsets,connection):
        self._modelsets=modelsets
        self._connection=connection
        self.__getfreqparams()

    def __getfreqparams(self):
        if isinstance(self._modelsets[0], company_modelset):
            if (self._modelsets[0]._mparams['ttm']):
                self._freq = AFREQS[self._modelsets[0]._mparams['reportdate'][-4:]]
                self._AP = 1
            else:
                self._freq = 'Q-DEC'
                self._AP = 4
        else:
            raise Exception("模型类型错误")
        reportlist = []
        for modelset in self._modelsets:
            if isinstance(modelset, company_modelset):
                if (modelset._mparams['ttm']):
                    if (AFREQS[modelset._mparams['reportdate'][-4:]] == self._freq):
                        reportlist.append(pd.to_datetime(modelset._mparams['reportdate']))
                    else:
                        raise Exception("模型类型错误")
                else:
                    reportlist.append(pd.to_datetime(modelset._mparams['reportdate']))
            else:
                raise Exception("模型类型错误")
        reportmax = np.array(reportlist).max()
        periodgaps = []
        for report in reportlist:
            periodgaps.append(len(pd.date_range(start=report, end=reportmax, freq=self._freq)) - 1)
        self._periodgaps = periodgaps

    def irr_trade(self,dimension,pricemode,tradedate,droplow=0):
        rates=[]
        mnum = len(self._modelsets)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        for i in range(mnum):
            irr=self._modelsets[i].irr_trade(dimension,pricemode,tradedate)
            if type(irr)==float:
                if self._modelsets[i]._mparams['seasonal']:
                    rates.append(1+irr)
                else:
                    rates.append(pow(1 + irr,self._AP))
        if len(rates)==0:
            return None
        elif len(rates)==1:
            return rates[0]-1
        rates=np.log(pd.Series(rates))
        resonances=[]
        for i in range(len(rates)):
            cw=0
            for j in range(len(rates)):
                if i==j:
                    continue
                cw+=self._connection.elementmethod(rates[i],rates[j])
            if cw==0:
                return np.exp(rates[0])-1
            resonances.append(1/cw)
        resonances=pd.Series(resonances)
        for m in range(droplow-(mnum-len(rates))):
            dinx=resonances.idxmin()
            del resonances[dinx]
            del rates[dinx]
        if len(rates)>0:
            return np.exp((rates * resonances).sum() / (resonances.sum()))-1
        else:
            return None

    def batchirr_trade(self,dimension,pricemode,startdate,enddate,droplow=0):
        batchrates=[]
        mnum = len(self._modelsets)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        for i in range(mnum):
            if self._modelsets[i]._mparams['seasonal']:
                batchrates.append(1+self._modelsets[i].batchirr_trade(dimension,pricemode,startdate,enddate)['IRRs'])
            else:
                batchrates.append(pow(1+self._modelsets[i].batchirr_trade(dimension,pricemode,startdate,enddate)['IRRs'],self._AP))
        batchrates=np.log(pd.DataFrame(batchrates).transpose())
        irrs=pd.Series(index=batchrates.index)
        for index,row in batchrates.iterrows():
            rates=row.copy(deep=True).dropna()
            if len(rates) == 0:
                irrs[index] = np.nan
                continue
            elif len(rates) == 1:
                irrs[index] = np.exp(rates.values[0]) - 1
                continue
            rates=rates.reset_index(drop=True)
            resonances = []
            homo = False
            for i in range(len(rates)):
                cw = 0
                for j in range(len(rates)):
                    if i == j:
                        continue
                    cw += self._connection.elementmethod(rates[i], rates[j])
                if cw == 0:
                    homo = True
                    break
                resonances.append(1 / cw)
            if homo:
                resonances = pd.Series(np.ones(len(rates)))
            else:
                resonances = pd.Series(resonances)
            for m in range(droplow-(mnum-len(rates))):
                dinx = resonances.idxmin()
                del resonances[dinx]
                del rates[dinx]
            if len(rates)>0:
                irrs[index] = np.exp((rates * resonances).sum() / (resonances.sum()))-1
            else:
                irrs[index] = np.nan
        return irrs

    def achieverate_trade(self,dimension, pricemode, tradedate, r,droplow=0):
        acrs = []
        mnum = len(self._modelsets)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        sr = pow(1 + r, 1 / self._AP) - 1
        for i in range(mnum):
            if self._modelsets[i]._mparams['seasonal']:
                acrs.append(self._modelsets[i].achieverate_trade(dimension,pricemode, tradedate, r))
            else:
                acrs.append(self._modelsets[i].achieverate_trade(dimension,pricemode, tradedate, sr))
        if len(acrs) == 0:
            return None
        elif len(acrs) == 1:
            return acrs[0]
        acrs = pd.Series(acrs)
        resonances = []
        for i in range(mnum):
            cw = 0
            for j in range(mnum):
                if i == j:
                    continue
                cw += self._connection.elementmethod(acrs[i], acrs[j])
            if cw == 0:
                return acrs[0]
            resonances.append(1 / cw)
        resonances = pd.Series(resonances)
        for m in range(droplow):
            dinx=resonances.idxmin()
            del resonances[dinx]
            del acrs[dinx]
        return (acrs*resonances).sum()/(resonances.sum())

    def batchachieverate_trade(self,dimension, pricemode, startdate, enddate, r,droplow=0):
        batchacrs=[]
        mnum = len(self._modelsets)
        if droplow>=mnum:
            raise Exception("异常剔除数量应小于分析模型数量")
        sr = pow(1 + r, 1 / self._AP) - 1
        for i in range(mnum):
            if self._modelsets[i]._mparams['seasonal']:
                batchacrs.append(self._modelsets[i].batchachieverate_trade(dimension,pricemode, startdate, enddate, r))
            else:
                batchacrs.append(self._modelsets[i].batchachieverate_trade(dimension,pricemode, startdate, enddate,sr))
        batchacrs=pd.DataFrame(batchacrs).transpose()
        acrs=pd.Series(index=batchacrs.index)
        for index,row in batchacrs.iterrows():
            rates=row.copy(deep=True)
            if len(rates) == 0:
                acrs[index]=np.nan
                continue
            elif len(rates) == 1:
                acrs[index]=rates.values[0]
                continue
            resonances = []
            homo = False
            for i in range(mnum):
                cw = 0
                for j in range(mnum):
                    if i == j:
                        continue
                    cw += self._connection.elementmethod(rates[i], rates[j])
                if cw == 0:
                    homo = True
                    break
                resonances.append(1 / cw)
            if homo:
                resonances = pd.Series(np.ones(len(rates)))
            else:
                resonances = pd.Series(resonances)
            for m in range(droplow):
                dinx = resonances.idxmin()
                del resonances[dinx]
                del rates[dinx]
            acrs[index] = (rates*resonances).sum()/(resonances.sum())
        return acrs





# 著作者：叶凯文
# Author: YE KAIWEN
# 版权归著作者叶凯文所有
# Copyright (c) 2022 YE KAIWEN. All Rights Reserved.

import requests
import json

vqdomain='https://valuequant.boomevolve.com/'

class VQconnect:
    def login(self,name,pswd):
        url=vqdomain+'login'
        params={'name':name,'pswd':pswd}
        try:
            res=requests.post(url=url,data=params)
            resInfo = json.loads(res.text)
            if res.status_code==200:
                self.token=resInfo['token']
                print(resInfo['msg'])
            else:
                print(resInfo['msg'])
        except:
            print("登录出错！")

    def logout(self):
        try:
            self.tokencheck()
            url = vqdomain + 'logout'
            params={}
            res = requests.post(url=url, data=params, headers={'BOOM-TOKEN': self.token})
            resInfo = json.loads(res.text)
            if res.status_code == 200:
                print(resInfo['msg'])
                del self.token
            else:
                print('注销登录失败')
        except:
            print('注销登录失败')

    def tokencheck(self):
        if hasattr(self,'token')==False:
            print("请先登录。若需注册，请登陆www.boomevolve.com。")
            raise Exception("请先登录。")

    def service(self,params):
        self.tokencheck()
        url=vqdomain+'service'
        try:
            params['params']=json.dumps(params['params'],ensure_ascii=False)
            res=requests.post(url=url,data=params,headers={'BOOM-TOKEN':self.token})
            resInfo=json.loads(res.text)
            if 'token' in resInfo.keys():
                self.token = resInfo['token']
            if res.status_code==200:
                return resInfo['res']
            else:
                print(resInfo['msg'])
                return None
        except Exception as e:
            print(e)
            print("请求错误")
            return None

    def data(self,params):
        self.tokencheck()
        url=vqdomain+'data'
        try:
            params['params'] = json.dumps(params['params'],ensure_ascii=False)
            res=requests.post(url=url,data=params,headers={'BOOM-TOKEN':self.token})
            resInfo=json.loads(res.text)
            if 'token' in resInfo.keys():
                self.token = resInfo['token']
            if res.status_code==200:
                return resInfo['res']
            else:
                print(resInfo['msg'])
                return None
        except:
            print("请求错误")
            return None

    def stocknet(self,params):
        self.tokencheck()
        url=vqdomain+'stocknet'
        try:
            params['params'] = json.dumps(params['params'],ensure_ascii=False)
            res=requests.post(url=url,data=params,headers={'BOOM-TOKEN':self.token})
            resInfo=json.loads(res.text)
            if 'token' in resInfo.keys():
                self.token = resInfo['token']
            if res.status_code==200:
                return resInfo['res']
            else:
                print(resInfo['msg'])
                return None
        except:
            print("请求错误")
            return None

VQC=VQconnect()


# 著作者：叶凯文
# Author: YE KAIWEN
# 版权归著作者叶凯文所有
# Copyright (c) 2022 YE KAIWEN. All Rights Reserved.

import pandas as pd

from .commonform import getcommonservice,seriestoformat
from .companyform import getcompanymodel, getcompanyservice, getcompanydata

AFREQS={'3-31':'A-MAR','6-30':'A-JUN','9-30':'A-SEP','2-31':'A-DEC'}

class company_modelset:

    def __init__(self,mparams):
        if mparams['adjdate']:
            pass
        else:
            mparams['adjdate']=mparams['reportdate']
        self._mparams = mparams
        self.__getmodelset()

    def resetparams(self, mparams):
        self._mparams = mparams
        if hasattr(self, '_modelset'):
            del self._modelset
        if hasattr(self, '_fitset'):
            del self._fitset
        if hasattr(self,'_incdata'):
            del self._incdata
        self.__getmodelset()

    def __getmodelset(self):
        res = getcompanymodel(mfunc=type(self).__name__, mparams=self._mparams.copy())
        if type(res) == dict:
            self._modelset = res['modelset']
            self._fitset=res['fitset']

    def modelset(self):
        if hasattr(self, '_modelset')==False:
            self.__getmodelset()
        if hasattr(self, '_modelset'):
            return self._modelset
        else:
            return None

    def fit(self,dimension):
        if hasattr(self, '_fitset')==False:
            self.__getmodelset()
        if hasattr(self, '_fitset'):
            if self._mparams['ttm']:
                inx = pd.date_range(end=self._mparams['reportdate'], freq=AFREQS[self._mparams['reportdate'][-4:]], periods=len(self._fitset[dimension]))
            else:
                inx=pd.date_range(end=self._mparams['reportdate'],freq='Q-DEC',periods=len(self._fitset[dimension]))
            if 'model' in self._modelset[dimension].keys():
                if 'step' in self._modelset[dimension]['model'].keys():
                    return pd.Series(self._fitset[dimension][-self._modelset[dimension]['model']['step']:],index=inx[-self._modelset[dimension]['model']['step']:]).dropna()
            return pd.Series(self._fitset[dimension], index=inx[-len(self._fitset[dimension]):]).dropna()
        else:
            return None

    def __getservice(self, restype, vparams):
        if hasattr(self, '_modelset') == False:
            self.__getmodelset()
        if hasattr(self, '_modelset'):
            res = getcommonservice(restype=restype, model=self._modelset, mfunc=type(self).__name__, vparams=vparams)
            return res
        else:
            return None

    def forecast(self,dimension, t):
        if self._mparams['ttm']:
            inx = pd.date_range(start=self._mparams['reportdate'], freq=AFREQS[self._mparams['reportdate'][-4:]], periods=(1 + t))
        else:
            inx = pd.date_range(start=self._mparams['reportdate'], freq='Q-DEC', periods=(1 + t))
        fore=self.__getservice(restype='Forecast', vparams={'t': t,'dimension':dimension})
        if type(fore)==list:
            return pd.Series(fore,index=inx[-t:])
        else:
            return None

    def pv(self,dimension, r):
        return self.__getservice(restype='PV', vparams={'r': r,'dimension':dimension})

    def __getdcdate(self):
        if self._mparams['ttm']:
            return pd.date_range(start=self._mparams['reportdate'], freq=AFREQS[self._mparams['reportdate'][-4:]], periods=2)[-1]
        else:
            return pd.date_range(start=self._mparams['reportdate'],freq='Q-DEC',periods=2)[-1]

    def __getrp(self):
        if self._mparams['ttm']:
            return 360
        elif self._mparams['seasonal']:
            return 360
        else:
            return 90

    def irr(self,dimension, price, tradedate):
        return self.__getservice(restype='UnitIRR', vparams={'price': price,'dimension':dimension,'rp':self.__getrp(), 'gap': (self.__getdcdate() - pd.to_datetime(tradedate)).days})

    def batchirr(self,dimension, priceseries):
        vparams = {}
        vparams['pricelist'], vparams['pricestart'], vparams['pricefreq'] = seriestoformat(priceseries)
        vparams['dimension']=dimension
        vparams['dcdate'] = self.__getdcdate().strftime('%Y-%m-%d')
        vparams['rp'] = self.__getrp()
        res=self.__getservice(restype='BatchIRR', vparams=vparams)
        if type(res)==dict:
            df=pd.DataFrame()
            df['IRRs']=pd.Series(res['IRRs'],index=priceseries.index)
            df['errors']=pd.Series(res['errors'],index=priceseries.index)
            return df
        else:
            return None

    def achieverate(self,dimension, price, r, tradedate):
        return self.__getservice(restype='UnitAchieverate', vparams={'price': price, 'r': r,'dimension':dimension,'rp':self.__getrp(), 'gap': (self.__getdcdate() - pd.to_datetime(tradedate)).days})

    def batchachieverate(self,dimension, priceseries, r):
        vparams = {}
        vparams['pricelist'], vparams['pricestart'], vparams['pricefreq'] = seriestoformat(priceseries)
        vparams['dimension']=dimension
        vparams['r']=r
        vparams['dcdate'] = self.__getdcdate().strftime('%Y-%m-%d')
        vparams['rp'] = self.__getrp()
        res=self.__getservice(restype='BatchAchieverate', vparams=vparams)
        if type(res)==list:
            return pd.Series(res,index=priceseries.index)
        else:
            return None

    def __gettradeservice(self, restype, vparams):
        if hasattr(self, '_modelset') == False:
            self.__getmodelset()
        if hasattr(self, '_modelset'):
            res = getcompanyservice(restype=restype, model=self._modelset, mparams=self._mparams.copy(), mfunc=type(self).__name__, vparams=vparams)
            return res
        else:
            return None

    def irr_trade(self,dimension, pricemode, tradedate):
        vparams = locals()
        del vparams['self']
        vparams['rp'] = self.__getrp()
        return self.__gettradeservice(restype='UnitIRR', vparams=vparams)

    def batchirr_trade(self,dimension, pricemode, startdate, enddate):
        vparams = locals()
        del vparams['self']
        vparams['rp'] = self.__getrp()
        res=self.__gettradeservice(restype='BatchIRR', vparams=vparams)
        if type(res)==dict:
            df=pd.DataFrame()
            df['IRRs']=pd.Series(res['IRRs'])
            df['errors']=pd.Series(res['errors'])
            return df
        else:
            return None

    def achieverate_trade(self,dimension, pricemode, tradedate, r):
        vparams = locals()
        del vparams['self']
        vparams['rp'] = self.__getrp()
        return self.__gettradeservice(restype='UnitAchieverate', vparams=vparams)

    def batchachieverate_trade(self,dimension, pricemode, startdate, enddate, r):
        vparams = locals()
        del vparams['self']
        vparams['rp'] = self.__getrp()
        res=self.__gettradeservice(restype='BatchAchieverate', vparams=vparams)
        if type(res)==dict:
            return pd.Series(res)
        else:
            return None

    def __getincdata(self,dimension):
        res = getcompanydata(stockcode=self._mparams['stockcode'], reportdate=self._mparams['reportdate'], dimension=dimension, adjdate=self._mparams['adjdate'],ttm=self._mparams['ttm'])
        if type(res) == dict:
            data=pd.Series(res)
            data.index=pd.to_datetime(data.index)
            self._incdata[dimension]=data

    def incdata(self,dimension):
        if hasattr(self,'_incdata')==False:
            self._incdata={}
            self.__getincdata(dimension)
        else:
            if dimension not in self._incdata.keys():
                self.__getincdata(dimension)
        if dimension in self._incdata.keys():
            return self._incdata[dimension]
        else:
            return None

class strict_latest_status_modelset(company_modelset):

    def __init__(self, stockcode,structure,reportdate, minstep, significance, seasonal=False, anneal=None, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().__init__(mparams)

    def resetparams(self, stockcode,structure,reportdate, minstep, significance, seasonal=False, anneal=None, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().resetparams(mparams)

class probable_latest_status_modelset(company_modelset):

    def __init__(self, stockcode,structure,reportdate, minstep, significance, seasonal=False, anneal=None, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().__init__(mparams)

    def resetparams(self, stockcode,structure,reportdate, minstep, significance, seasonal=False, anneal=None, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().resetparams(mparams)

class long_range_status_modelset(company_modelset):

    def __init__(self, stockcode,structure,reportdate, significance, seasonal=False, anneal=None, samplerange=None, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().__init__(mparams)

    def resetparams(self,stockcode,structure,reportdate, significance, seasonal=False, anneal=None, samplerange=None, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().resetparams(mparams)

class strict_latest_long_range_modelset(company_modelset):

    def __init__(self,stockcode,structure,reportdate, minstep, significance, inertia, seasonal=False, anneal=None, samplerange=None, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().__init__(mparams)

    def resetparams(self,stockcode,structure,reportdate, minstep, significance, inertia, seasonal=False, anneal=None, samplerange=None, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().resetparams(mparams)

class probable_latest_long_range_modelset(company_modelset):

    def __init__(self, stockcode,structure,reportdate, minstep, significance, inertia, seasonal=False, anneal=None, samplerange=None, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().__init__(mparams)

    def resetparams(self, stockcode,structure,reportdate, minstep, significance, inertia, seasonal=False, anneal=None, samplerange=None, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().resetparams(mparams)

class series_shape_modelset(company_modelset):

    def __init__(self,stockcode,structure,reportdate, significance, seasonal=False, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().__init__(mparams)

    def resetparams(self,stockcode,structure,reportdate, significance, seasonal=False, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().resetparams(mparams)

class ma_strict_latest_status_shape_modelset(company_modelset):

    def __init__(self,stockcode,structure,reportdate,mawindow,minstep, significance, seasonal=False, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().__init__(mparams)

    def resetparams(self,stockcode,structure,reportdate,mawindow,minstep, significance, seasonal=False, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().resetparams(mparams)

class ma_probable_latest_status_shape_modelset(company_modelset):

    def __init__(self,stockcode,structure,reportdate,mawindow,minstep, significance, seasonal=False, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().__init__(mparams)

    def resetparams(self,stockcode,structure,reportdate,mawindow,minstep, significance, seasonal=False, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().resetparams(mparams)

class ma_strict_latest_status_modelset(company_modelset):

    def __init__(self,stockcode,structure,reportdate, mawindow, minstep, seasonal=False, anneal=None, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().__init__(mparams)

    def resetparams(self,stockcode,structure,reportdate, mawindow, minstep, seasonal=False, anneal=None, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().resetparams(mparams)

class ma_probable_latest_status_modelset(company_modelset):

    def __init__(self,stockcode,structure,reportdate, mawindow, minstep,significance, seasonal=False, anneal=None, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().__init__(mparams)

    def resetparams(self,stockcode,structure,reportdate, mawindow, minstep,significance, seasonal=False, anneal=None, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().resetparams(mparams)

class ma_long_range_status_modelset(company_modelset):

    def __init__(self, stockcode,structure,reportdate, mawindow, significance, seasonal=False, anneal=None, samplerange=None, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().__init__(mparams)

    def resetparams(self,stockcode,structure,reportdate, mawindow, significance, seasonal=False, anneal=None, samplerange=None, conservative=False, adjdate=None, ttm=False):
        mparams = locals()
        del mparams['self']
        del mparams['__class__']
        super().resetparams(mparams)


# 著作者：叶凯文
# Author: YE KAIWEN
# 版权归著作者叶凯文所有
# Copyright (c) 2022 YE KAIWEN. All Rights Reserved.

from .VQconnect import VQC

import pandas as pd

datasize=5000
accuracy=4

modelcate={
    'strict_latest_status_model':'Status',
    'probable_latest_status_model':'Status',
    'long_range_status_model':'Status',
    'strict_latest_long_range_model':'Status',
    'probable_latest_long_range_model':'Status',
    'series_shape_model':'Shape',
    'ma_strict_latest_status_shape_model':'Shape',
    'ma_probable_latest_status_shape_model':'Shape',
    'ma_strict_latest_status_model':'Ma',
    'ma_probable_latest_status_model':'Ma',
    'ma_long_range_status_model':'Ma',
}

def formatseries(datalist,startdate,freq):
    if len(datalist)>5000:
        print("上传数据序列长度超过5000。")
        return None
    try:
        series=pd.Series(datalist,index=pd.date_range(start=startdate,freq=freq,periods=len(datalist)))
        if series.dtypes not in [int,float]:
            print("序列元素应为数字类型。")
            return None
        return series.round(accuracy)
    except:
        print("数据出错,请检查序列参数。")
        return None

def formatdataframe(datalist,startdate,freq,datacols):
    try:
        df=pd.DataFrame(datalist,index=pd.date_range(start=startdate,freq=freq,periods=len(datalist)),columns=datacols)
        if (len(df.index)*len(df.columns))>datasize:
            print("上传数据表格元素总数超过5000。")
            return None
        if (((df.dtypes==int).all())|((df.dtypes==float).all()))==False:
            print("数据表格元素应为数字类型。")
            return None
        return df.round(accuracy)
    except:
        print("数据出错,请检查数据表格参数。")
        return None

def seriestoformat(series):
    datalist = series.values.tolist()
    startdate = series.index[0].strftime('%Y-%m-%d')
    freq = series.index.freq.name
    return datalist,startdate,freq

def dftoformat(df):
    datalist = df.values.tolist()
    startdate = df.index[0].strftime('%Y-%m-%d')
    freq = df.index.freq.name
    datacols = df.columns.to_list()
    return datalist,startdate,freq,datacols

def formatmodelparams(mfunc,mparams):
    datalist=mparams.pop('datalist')
    startdate=mparams.pop('startdate')
    freq=mparams.pop('freq')
    series=formatseries(datalist,startdate,freq)
    if type(series)==pd.Series:
        return {'datalist':datalist,'startdate':startdate,'freq':freq,'mcate':modelcate[mfunc],'mfunc':mfunc,'mparams':mparams}
    else:
        return None

def formatmodelsetparams(mfunc,mparams):
    datalist=mparams.pop('datalist')
    startdate=mparams.pop('startdate')
    freq=mparams.pop('freq')
    datacols=mparams.pop('datacols')
    structure=mparams.pop('structure')
    mfunc=mfunc[:-3]
    if type(structure) != dict:
        print('ModelSet结构错误。')
        return None
    df = formatdataframe(datalist, startdate, freq, datacols)
    if type(df)==pd.DataFrame:
        return {'datalist':datalist,'startdate':startdate,'freq':freq,'datacols':datacols,'structure':structure,'mcate':modelcate[mfunc],'mfunc':mfunc,'mparams':mparams}
    else:
        return None

def getcommonmodel(mfunc,mparams):
    modeltype=mfunc.split('_')[-1]
    if modeltype=='model':
        modeltype='Model'
    elif modeltype=='modelset':
        modeltype='ModelSet'
    else:
        print("非法参数")
        return None
    formatparams=globals()['format'+modeltype.lower()+'params'](mfunc,mparams)
    if type(formatparams)==dict:
        try:
            params = {'datatype': 'Common', 'modeltype': modeltype, 'restype': 'Model', 'params':formatparams}
            return VQC.service(params)
        except:
            print("网络请求出错")
            return None
    else:
        print("非法参数")
        return None

def formatmodelxparams(model,mfunc):
    if type(model)==dict:
        return {'model':model,'mcate':modelcate[mfunc]}
    else:
        return None

def formatmodelsetxparams(modelset,mfunc,vparams):
    if type(modelset)==dict:
        return {'modelset':modelset,'mcate':modelcate[(mfunc[:-8]+'model')],'dimension':vparams.pop('dimension')}
    else:
        return None

def formatdimensionparams(fparams,vparams):
    if type(fparams)==dict:
        fparams['dimension']=vparams.pop('dimension')
        return fparams
    else:
        return fparams

def formatforecastparams(fparams,vparams):
    if type(fparams)==dict:
        fparams['t']=vparams.pop('t')
        return fparams
    else:
        return fparams

def formatpvparams(fparams,vparams):
    if type(fparams)==dict:
        fparams['r']=vparams.pop('r')
        return fparams
    else:
        return fparams

def formatbatchirrparams(fparams,vparams):
    if type(fparams)==dict:
        pricelist=vparams.pop('pricelist')
        pricestart=vparams.pop('pricestart')
        pricefreq = vparams.pop('pricefreq')
        series=formatseries(pricelist,pricestart,pricefreq)
        if type(series)==pd.Series:
            fparams['pricelist']=pricelist
            fparams['pricestart'] = pricestart
            fparams['pricefreq'] = pricefreq
            fparams['dcdate'] = vparams.pop('dcdate')
            fparams['rp']=vparams.pop('rp')
            return fparams
        else:
            return None
    else:
        return fparams

def formatunitirrparams(fparams,vparams):
    if type(fparams)==dict:
        fparams['price'] = vparams.pop('price')
        fparams['gap']=vparams.pop('gap')
        fparams['rp']=vparams.pop('rp')
        return fparams
    else:
        return fparams

def formatbatchachieverateparams(fparams,vparams):
    if type(fparams)==dict:
        pricelist=vparams.pop('pricelist')
        pricestart=vparams.pop('pricestart')
        pricefreq = vparams.pop('pricefreq')
        series=formatseries(pricelist,pricestart,pricefreq)
        if type(series)==pd.Series:
            fparams['pricelist']=pricelist
            fparams['pricestart'] = pricestart
            fparams['pricefreq'] = pricefreq
            fparams['r']=vparams.pop('r')
            fparams['dcdate'] = vparams.pop('dcdate')
            fparams['rp'] = vparams.pop('rp')
            return fparams
        else:
            return None
    return fparams

def formatunitachieverateparams(fparams,vparams):
    if type(fparams)==dict:
        fparams['price'] = vparams.pop('price')
        fparams['gap']=vparams.pop('gap')
        fparams['r'] = vparams.pop('r')
        fparams['rp']=vparams.pop('rp')
        return fparams
    else:
        return fparams

def getcommonservice(restype,model,mfunc,vparams):
    modeltype=mfunc.split('_')[-1]
    if modeltype=='model':
        fparams=formatmodelxparams(model=model,mfunc=mfunc)
        modeltype='Model'
    elif modeltype in ['modelset','frameset']:
        fparams=formatmodelsetxparams(modelset=model,mfunc=mfunc,vparams=vparams)
        modeltype='ModelSet'
    else:
        print("非法参数")
        return None
    fparams=globals()['format'+restype.lower()+'params'](fparams=fparams,vparams=vparams)
    if type(fparams)==dict:
        try:
            params = {'datatype': 'Common', 'modeltype': modeltype, 'restype': restype, 'params':fparams}
            return VQC.service(params)
        except:
            print("网络请求出错")
            return None
    else:
        return None



# 著作者：叶凯文
# Author: YE KAIWEN
# 版权归著作者叶凯文所有
# Copyright (c) 2022 YE KAIWEN. All Rights Reserved.

import pandas as pd

from .VQconnect import VQC

class StockNet:
    def __init__(self,freq,periodend,analysiscate,dimension):
        self._netparams={'freq':freq,'periodend':periodend,'analysiscate':analysiscate,'dimension':dimension}
        self.__getNetworkInfo()

    def __getNetworkInfo(self):
        params={'querytype':'NetworkInfo','params':self._netparams}
        try:
            res=VQC.stocknet(params)
            if type(res)==list:
                self._NetID=res[0]
                self._NetworkInfo={'vertexnum':res[1],'edgenum':res[2],'simulimited':res[3]}
        except:
            print("网络请求出错")
            pass

    def NetworkInfo(self):
        if hasattr(self,'_NetworkInfo'):
            return self._NetworkInfo

    def __getNetworkGraphs(self,radius):
        params={'querytype':'NetworkGraphs','params':{'nid':self._NetID,'radius':radius}}
        try:
            res=VQC.stocknet(params)
            if type(res)==list:
                Graphs=pd.DataFrame(res,columns=['sid','vertexnum','edgenum']).set_index('sid',drop=True)
                if hasattr(self,'_Graphs'):
                    self._Graphs[radius]=Graphs
                else:
                    self._Graphs={radius:Graphs}
        except:
            print("网络请求出错")
            pass

    def NetworkGraphs(self,radius):
        if hasattr(self,'_Graphs'):
            if radius in self._Graphs.keys():
                return self._Graphs[radius]
        self.__getNetworkGraphs(radius)
        if hasattr(self,'_Graphs'):
            if radius in self._Graphs.keys():
                return self._Graphs[radius]

    def __getGraphVertexes(self,radius,sid):
        params={'querytype':'GraphVertexes','params':{'nid':self._NetID,'radius':radius,'sid':sid}}
        try:
            res = VQC.stocknet(params)
            if type(res) == list:
                Vertexes=res
                if hasattr(self,'_Vertexes'):
                    self._Vertexes['%s_%s'%(radius,sid)]=Vertexes
                else:
                    self._Vertexes={'%s_%s'%(radius,sid):Vertexes}
        except:
            print("网络请求出错")
            pass

    def GraphVertexes(self,radius,sid):
        if hasattr(self,'_Vertexes'):
            if '%s_%s'%(radius,sid) in self._Vertexes.keys():
                return self._Vertexes['%s_%s'%(radius,sid)]
        self.__getGraphVertexes(radius,sid)
        if hasattr(self,'_Vertexes'):
            if '%s_%s'%(radius,sid) in self._Vertexes.keys():
                return self._Vertexes['%s_%s'%(radius,sid)]

    def VertexNeighbors(self,vertex,edgemode='out'):
        params={'querytype':'VertexNeighbors','params':{'nid':self._NetID,'vertex':vertex,'edgemode':edgemode}}
        try:
            res=VQC.stocknet(params)
            if type(res) == list:
                if edgemode=='out':
                    return pd.DataFrame(res,columns=['endvertex','simularity'])
                elif edgemode=='in':
                    return pd.DataFrame(res,columns=['startvertex','simularity'])
                else:
                    print("edgemode应选择'in'或'out'其中一项")
        except:
            print("网络请求出错")
            pass

    def EdgeFrequency(self,startvertex,endvertex,simulimited):
        params={'querytype':'EdgeFrequency','params':{'startvertex':startvertex,'endvertex':endvertex,'simulimited':simulimited,'freq':self._netparams['freq'],'analysiscate':self._netparams['analysiscate'],'dimension':self._netparams['dimension']}}
        try:
            res=VQC.stocknet(params)
            if type(res)==list:
                return res
        except:
            pass
        print("网络请求出错")

class ParticleCompetitionTraining:
    def __init__(self,network,sid):
        self._trainparams={'nid':network._NetID,'sid':sid}
        self.__getTrainingInfo()

    def __getTrainingInfo(self):
        params={'querytype':'ParticleCompetitionTrainingInfo','params':{'nid':self._trainparams['nid'],'sid':self._trainparams['sid']}}
        try:
            res = VQC.stocknet(params)
            if type(res)==list:
                self._TrainingInfo=pd.DataFrame(res,columns=['tid','psize','avemaxctl']).set_index('tid',drop=True)
                self._TrainingInfo.index=self._TrainingInfo.index.astype(int)
                self._TrainingInfo['psize']=self._TrainingInfo['psize'].astype(int)
        except:
            print("网络请求出错")
            pass

    def TrainingInfo(self):
        if hasattr(self,'_TrainingInfo'):
            return self._TrainingInfo
        self.__getTrainingInfo()
        if hasattr(self,'_TrainingInfo'):
            return self._TrainingInfo

    def __getTrainingParticles(self,tid):
        params={'querytype':'TrainingParticles','params':{'tid':tid}}
        try:
            res=VQC.stocknet(params)
            if type(res)==list:
                Particles=pd.DataFrame(res,columns=['pid','monopolize','overlap','competeL1','competeL2']).set_index('pid',drop=True)
                if hasattr(self,'_Particles'):
                    self._Particles[tid]=Particles
                else:
                    self._Particles={tid:Particles}
        except:
            print("网络请求出错")
            pass

    def Particles(self,tid):
        if hasattr(self, '_Particles'):
            if tid in self._Particles.keys():
                return self._Particles[tid]
        self.__getTrainingParticles(tid)
        if hasattr(self, '_Particles'):
            if tid in self._Particles.keys():
                return self._Particles[tid]

    def __getTrainingVertexes(self,tid):
        params={'querytype':'TrainingVertexes','params':{'tid':tid}}
        try:
            res=VQC.stocknet(params)
            if type(res)==list:
                Vertexes=pd.DataFrame(res,columns=['vertex','uv','pv','top1','top1freq','top2','top2freq','top3','top3freq','top4','top4freq','top5','top5freq','overlap','maxctlnum'])
                if hasattr(self,'_Vertexes'):
                    self._Vertexes[tid]=Vertexes
                else:
                    self._Vertexes={tid:Vertexes}
        except:
            print("网络请求出错")
            pass

    def Vertexes(self,tid):
        if hasattr(self,'_Vertexes'):
            if tid in self._Vertexes.keys():
                return self._Vertexes[tid]
        self.__getTrainingVertexes(tid)
        if hasattr(self,'_Vertexes'):
            if tid in self._Vertexes.keys():
                return self._Vertexes[tid]



# 著作者：叶凯文
# Author: YE KAIWEN
# 版权归著作者叶凯文所有
# Copyright (c) 2022 YE KAIWEN. All Rights Reserved.

from setuptools import setup,find_packages

setup(
    name='valuequant',
    version='1.1.3',
    description='Provide stock value modeling and market factor analysis service.',
    long_description='valuequant provide stock value modeling and market factor analysis service. Register for a service account and get start on https://www.boomevolve.com',
    long_description_content_type='text/plain',
    url='https://www.boomevolve.com/valuequant',
    author='YE KAIWEN',
    author_email='kevinye@originalbooms.com',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    keywords=("Quant","Data Science","Statistics","Value Analysis"),
    project_urls={
        'Documentation':'https://www.boomevolve.com/valuequant',
        'Register':'https://www.boomevolve.com/'
    },
    packages=find_packages(include=['valuequant','valuequant.*']),
    install_requires=['pandas','requests'],
    python_requires='>=3',
)



# 著作者：叶凯文
# Author: YE KAIWEN
# 版权归著作者叶凯文所有
# Copyright (c) 2022 YE KAIWEN. All Rights Reserved.

from valuequant.VQconnect import VQC
import valuequant.Models as Models
import valuequant.ModelSets as ModelSets
import valuequant.StockModels as StockModels
import valuequant.StockModelSets as StockModelSets
import valuequant.StockFrameSets as StockFrameSets
import valuequant.StockNet as StockNet
import valuequant.ResonanceModel as ResonanceModel

def login(name,pswd):
    VQC.login(name,pswd)

def logout():
    VQC.logout()


